/* 
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */

/*
 *  ======== SystemCfg_host.c ========
 *
 */

/* this define must precede inclusion of any xdc header file */
#define Registry_CURDESC SystemCfg__Desc
#define MODULE_NAME "SystemCfg"

#include <string.h>

/* xdctools header files */
#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Error.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Main.h>
#include <xdc/runtime/Registry.h>
#include <xdc/runtime/System.h>
#include <xdc/runtime/knl/SemThread.h>
#include <xdc/runtime/knl/Thread.h>

/* package header files */
#include <ti/ipc/HeapBufMP.h>
#include <ti/ipc/MessageQ.h>
#include <ti/ipc/MultiProc.h>
#include <ti/ipc/Notify.h>
#include <ti/syslink/IpcHost.h>
#include <ti/syslink/ProcMgr.h>
#include <ti/syslink/SysLink.h>

/* interface definition for this module */
#include <ti/sdo/rcm/examples/appcfg/SystemCfg.h>

/* local header files */
#include "Global.h"

/* instance object */
typedef struct SystemCfg_Object_tag {
    String              remoteProcName;
    ProcMgr_Handle      procMgrH;
    UInt32              fileId;
    SemThread_Struct    semObj;
    SemThread_Handle    semH;
    UInt16              remoteProcId;
    HeapBufMP_Handle    heapH;
} SystemCfg_Object;

static Char *procState[] = {
    "Unknown", "Powered", "Reset", "Loaded", "Running"
};

static Int curInit = 0;

/* module diags mask */
Registry_Desc Registry_CURDESC;


/*
 *  ======== SystemCfg_init ========
 */
Void SystemCfg_init(Void)
{
    Registry_Result result;


    if (curInit++ != 0) {
        return;  /* already initialized */
    }

    /* register with xdc.runtime to get a diags mask */
    result = Registry_addModule(&Registry_CURDESC, MODULE_NAME);
    Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);
}


/*
 *  ======== SystemCfg_exit ========
 */
Void SystemCfg_exit(Void)
{
//  Registry_Result result;


    if (curInit-- != 1) {
        return;  /* object still being used */
    }

    /* unregister from xdc.runtime */
//  result = Registry_removeModule(MODULE_NAME);
//  Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);
}


/*
 *  ======== SystemCfg_Params_init ========
 */
Void SystemCfg_Params_init(SystemCfg_Params *params)
{
    params->remoteProcName = NULL;
}


/*
 *  ======== SystemCfg_create ========
 */
#define FXNN "SystemCfg_create"
Int SystemCfg_create(const SystemCfg_Params *params,
        SystemCfg_Handle *handlePtr)
{
    Int status;
    Int n, bufSize;
    SystemCfg_Object *obj;
    Char *cp;
    Error_Block eb;


    Log_print0(Diags_ENTRY, "--> "FXNN": ()");

    /* initialize local vars */
    status = 0;
    Error_init(&eb);
    *handlePtr = (SystemCfg_Handle)NULL;

    /* allocate the object */
    obj = (SystemCfg_Handle)xdc_runtime_Memory_calloc(NULL,
        sizeof(SystemCfg_Object), sizeof(Int), &eb);

    if (obj == NULL) {
        Log_error1(FXNN": out of memory: size=%u", sizeof(SystemCfg_Object));
        status = SystemCfg_E_NOMEMORY;
        goto leave;
    }

    /* object-specific initialization */
    obj->procMgrH = NULL;
    obj->fileId = 0xFFFFFFFF;
    obj->semH = NULL;
    obj->remoteProcId = MultiProc_INVALIDID;
    obj->heapH = NULL;
    
    /* initialize SemThread_Struct to zero */
    for (cp = (Char *)&obj->semObj, n = sizeof(SemThread_Struct); n > 0; n--) {
        *cp++ = 0;
    }

    /* store the remote processor name */
    bufSize = strlen(params->remoteProcName) + 1;
    obj->remoteProcName = (String)xdc_runtime_Memory_calloc(NULL,
        bufSize, sizeof(String), &eb);

    if (obj == NULL) {
        Log_error1(FXNN": out of memory: size=%u", bufSize);
        status = SystemCfg_E_NOMEMORY;
        goto leave;
    }

    strcpy(obj->remoteProcName, params->remoteProcName);

    /* success, return opaque pointer */
    *handlePtr = (SystemCfg_Handle)obj;


leave:
    Log_print1(Diags_EXIT, "<-- "FXNN": %d", (IArg)status);
    return(status);
}
#undef FXNN


/*
 *  ======== SystemCfg_delete ========
 */
#define FXNN "SystemCfg_delete"
Int SystemCfg_delete(SystemCfg_Object **objPtr)
{
    Int bufSize;


    Log_print0(Diags_ENTRY, "--> "FXNN": ()");

    /* free the name buffer */
    bufSize = strlen((*objPtr)->remoteProcName) + 1;
    xdc_runtime_Memory_free(NULL, (Ptr)((*objPtr)->remoteProcName), bufSize);

    /* free the object memory */
    xdc_runtime_Memory_free(NULL, (Ptr)(*objPtr), sizeof(SystemCfg_Object));
    *objPtr = NULL;

    Log_print1(Diags_EXIT, "<-- "FXNN": %d", (IArg)SystemCfg_S_SUCCESS);
    return(SystemCfg_S_SUCCESS);
}
#undef FXNN


/*
 *  ======== SystemCfg_start ========
 */
#define FXNN "SystemCfg_start"
Int SystemCfg_start(SystemCfg_Object *obj, String filename)
{
    ProcMgr_AttachParams attachP;
    ProcMgr_StartParams startP;
    ProcMgr_State state;
    Int argc;
    String argv[2];
    Int status = 0;


    Log_print1(Diags_ENTRY, "--> "FXNN": (object: 0x%x)", (IArg)obj);

    /* SysLink initialize */
    Log_print0(Diags_INFO, FXNN": SysLink_setup");

    SysLink_setup();

    /* get the dsp procId, must be called *after* SysLink_setup() */
    obj->remoteProcId = MultiProc_getId(obj->remoteProcName);

    if (MultiProc_INVALIDID == obj->remoteProcId) {
        Log_error0(FXNN": MultiProc_getId() returned INVALID");
        status = -1;
        goto leave;
    }

    /* open a ProcMgr handle to the dsp */
    Log_print1(Diags_INFO,
        FXNN": ProcMgr_open: (remoteProcId: %d)",
        (IArg)obj->remoteProcId);

    status = ProcMgr_open(&obj->procMgrH, obj->remoteProcId);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_open() failed, %d", (IArg)status);
        goto leave;
    }

    /* attach to the remote processor */
    Log_print0(Diags_INFO, FXNN": ProcMgr_attach: (...)");

    ProcMgr_getAttachParams(NULL, &attachP);
    status = ProcMgr_attach(obj->procMgrH, &attachP);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_attach() failed, %d", (IArg)status);
        goto leave;
    }

    /* setup argc & argv for remote processor */
    argc = 1;
    argv[0] = filename;

    /* load and run the remote processor */
    Log_print2(Diags_INFO,
        FXNN": ProcMgr_load: (procId: %d, file: 0x%x)",
        (IArg)obj->remoteProcId, (IArg)filename);

    status = ProcMgr_load(obj->procMgrH, filename, argc, argv, NULL,
        &obj->fileId);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_load() failed, %d", (IArg)status);
        goto leave;
    }

    /* temporary: to be done by SysLink in the future */
    Log_print0(Diags_INFO, FXNN": Ipc_control: (LOADCALLBACK)");

    status = Ipc_control(obj->remoteProcId,
        Ipc_CONTROLCMD_LOADCALLBACK, NULL);

    if (status < 0) {
        Log_error1(FXNN": Ipc_control() failed, %d", (IArg)status);
        goto leave;
    }

    state = ProcMgr_getState(obj->procMgrH);
    Log_print1(Diags_INFO,
        FXNN": Processor state is %s", (IArg)procState[state]);

    /* start remote processor */
    Log_print0(Diags_INFO, FXNN": ProcMgr_start: (...)");

    ProcMgr_getStartParams(obj->procMgrH, &startP);
    status = ProcMgr_start(obj->procMgrH, &startP);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_start() failed, %d", (IArg)status);
        goto leave;
    }

    /* temporary: to be done by SysLink in the future */
    Log_print0(Diags_INFO, FXNN": Ipc_control: (STARTCALLBACK)");

    status = Ipc_control(obj->remoteProcId,
        Ipc_CONTROLCMD_STARTCALLBACK, NULL);

    if (status < 0) {
        Log_error1(FXNN": Ipc_control() failed, %d", (IArg)status);
        goto leave;
    }

    state = ProcMgr_getState(obj->procMgrH);
    Log_print1(Diags_INFO,
        FXNN": Processor state is %s", (IArg)procState[state]);


leave:
    status = (status >= 0 ? 0 : status);
    Log_print2(Diags_EXIT, "<-- %s: %d", (IArg)FXNN, (IArg)status);
    return(status);
}
#undef FXNN


/*
 *  ======== SystemCfg_stop ========
 */
#define FXNN "SystemCfg_stop"
Int SystemCfg_stop(SystemCfg_Object *obj)
{
    ProcMgr_State state;
    Int status;


    Log_print0(Diags_ENTRY, "--> "FXNN": ()");

    /* temporary, to be removed later */
    Log_print0(Diags_INFO, FXNN": Ipc_control: (STOPCALLBACK)");

    status = Ipc_control(obj->remoteProcId,
        Ipc_CONTROLCMD_STOPCALLBACK, NULL);

    if (status < 0) {
        Log_error1(FXNN": Ipc_control() failed, %d", (IArg)status);
        goto leave;
    }

    /* stop remote processor */
    Log_print0(Diags_INFO, FXNN": ProcMgr_stop: (...)");

    status = ProcMgr_stop(obj->procMgrH);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_stop() failed, %d", (IArg)status);
        goto leave;
    }

    /* unload the remote processor */
    Log_print0(Diags_INFO, FXNN": ProcMgr_unload: (...)");

    status = ProcMgr_unload(obj->procMgrH, obj->fileId);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_unload() failed, %d", (IArg)status);
        goto leave;
    }

    state = ProcMgr_getState(obj->procMgrH);
    Log_print1(Diags_INFO,
        FXNN": Processor state is %s", (IArg)procState[state]);

    /* detach from the remote processor */
    Log_print0(Diags_INFO, FXNN": ProcMgr_detach: (...)");

    status = ProcMgr_detach(obj->procMgrH);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_detach failed, %d", (IArg)status);
        goto leave;
    }

    /* close the ProcMgr object */
    Log_print0(Diags_INFO, FXNN": ProcMgr_close: (...)");

    status = ProcMgr_close(&obj->procMgrH);

    if (status < 0) {
        Log_error1(FXNN": ProcMgr_close() failed, %d", (IArg)status);
        goto leave;
    }

    /* SysLink finalize */
    Log_print0(Diags_INFO, FXNN": SysLink_destroy");

    SysLink_destroy();


leave:
    Log_print1(Diags_EXIT, "<-- "FXNN": %d", (IArg)status);
    return(status);
}
#undef FXNN


/*
 *  ======== SystemCfg_initializeResources ========
 */
#define FXNN "SystemCfg_initializeResources"
Int SystemCfg_initializeResources(SystemCfg_Object *obj, App_Fxn appStartupFxn,
    App_Handle arg)
{
    HeapBufMP_Params heapBufMPP;
    Int status = 0;


    Log_print0(Diags_ENTRY, "--> "FXNN": ()");

    /* create heap for messages */
    HeapBufMP_Params_init(&heapBufMPP);
    heapBufMPP.name = Global_RcmClientHeapName;
    heapBufMPP.regionId = 0;
    heapBufMPP.blockSize = 0x200;  /* 512 B */
    heapBufMPP.numBlocks = 8;

    obj->heapH = HeapBufMP_create(&heapBufMPP);

    if (obj->heapH == NULL) {
        Log_error0(FXNN": HeapBuf_create() failed");
        status = -1;
        goto leave;
    }

    /* register this heap with MessageQ */
    Log_print2(Diags_INFO,
        FXNN": MessageQ_registerHeap: (heapH: 0x%x, heapId: %d)",
        (IArg)(obj->heapH), (IArg)Global_RcmClientHeapId);

    MessageQ_registerHeap((Ptr)(obj->heapH), Global_RcmClientHeapId);

leave:
    Log_print1(Diags_EXIT, "<-- "FXNN": %d", (IArg)status);
    return(status);
}
#undef FXNN


/*
 *  ======== SystemCfg_finalizeResources ========
 */
#define FXNN "SystemCfg_finalizeResources"
Int SystemCfg_finalizeResources(SystemCfg_Object *obj, App_Fxn appShutdownFxn,
    App_Handle arg)
{
    Int status = 0;


    Log_print0(Diags_ENTRY, "--> "FXNN": ()");

    /* unregister heap with MessageQ */
    MessageQ_unregisterHeap(Global_RcmClientHeapId);

    /* delete heap used for rcm message queue */
    HeapBufMP_delete(&obj->heapH);

    Log_print1(Diags_EXIT, "<-- "FXNN": %d", (IArg)status);
    return(status);
}
#undef FXNN
/*
 *  @(#) ti.sdo.rcm.examples.hello; 1,0,0,38; 11-2-2010 18:30:36; /db/atree/library/trees/fc/fc-n22x/src/ xlibrary

 */

