/* 
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/*
 *  ======== dskt2free.c ========
 *  DSKT2 API to free the memory used by an XDAIS algorithm instance, and then
 *  delete the instance.
 *
 */

/* This define must precede inclusion of any xdc header files */
#define Registry_CURDESC ti_sdo_fc_dskt2_desc

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/Gate.h>
#include <xdc/runtime/Registry.h>

#include <ti/xdais/ialg.h>

#include <xdc/runtime/Memory.h>


#include "dskt2.h"
#include "_dskt2.h"

#include "dskt2_lock.h"

extern Registry_Desc ti_sdo_fc_dskt2_desc;

extern IGateProvider_Handle _DSKT2_gate;
extern Int _DSKT2_algReferenceCount;

extern Bool ti_sdo_fc_multicoreDsp;

#ifdef xdc_target__isaCompatible_64P
extern cregister volatile Uns DNUM;
extern Uns MC_ADDR_GLOB_TO_LOC(Uns glob_addr);
#else
#define MC_ADDR_GLOB_TO_LOC(addr) (addr)
#endif


/*
 *  ======== DSKT2_freeAlg ========
 */
Bool DSKT2_freeAlg(Int scratchMutexId, IALG_Handle alg)
{
    Int         numRecs;
    Int         memTabSize;
    IALG_MemRec *memTab;
    xdc_runtime_IHeap_Handle extHeap;
    Int         status = TRUE;
    IArg        key;
    Int         i = 0;

    Log_print2(Diags_ENTRY, "[+E] DSKT2_freeAlg> Enter "
            "(scratchMutexId=%d, alg=0x%x)", (IArg)scratchMutexId, (IArg)alg);

#ifndef DSKT2_DISABLELAZYD
    /*
     * If alg instance is currently active remove info from active table.
     * (No need to actually 'deactivate' the instance.)
     *
     * We need to do this early so pre-empting tasks know not to deactivate
     * alg.
     */
    if ((scratchMutexId >= 0) && (scratchMutexId < DSKT2_NUM_SCRATCH_GROUPS)) {
        key = IGateProvider_enter(_DSKT2_gate);

        if (_DSKT2_activeAlg[scratchMutexId] == alg) {
            _DSKT2_activeAlg[scratchMutexId] = NULL;
        }
        else if (_DSKT2_lastActiveAlg[scratchMutexId] == alg) {
            _DSKT2_lastActiveAlg[scratchMutexId] = NULL;
        }

        IGateProvider_leave(_DSKT2_gate, key);
    }
#endif

    /*
     *  Use the memTab stored during create phase to deallocate instance
     *  memory, as well as the memTab structure itself.
     *  No need to call alg's algFree function anymore.
     */
    Log_print0(Diags_USER2, "[+2] DSKT2_freeAlg> Dequeue alg information");

    memTab =_DSKT2_dequeueMemTab(_DSKT2_heap, alg, &memTabSize,
            &numRecs, &extHeap);

    if (memTab == NULL) {
        Log_print0(Diags_USER7, "[+7] DSKT2_freeAlg> Illegal Memory");

        /* return without freeing any memory */
        return (FALSE);
    }

    Log_print4(Diags_USER2, "[+2] DSKT2_freeAlg> Dequeued alg information, "
            "memTab 0x%x memTabSize %d, numRecs %d, extHeap 0x%x",
            (IArg)memTab, (IArg)memTabSize, (IArg)numRecs, (IArg)extHeap);

    for(i = 0; i < numRecs; i++) {

        //DEBUG TEMP PLEASE CHANGE
        if (ti_sdo_fc_multicoreDsp) {
            memTab[i].base = (void *)MC_ADDR_GLOB_TO_LOC((Uns)memTab[i].base);
        }
        //DEBUG TEMP PLEASE CHANGE
    }

    /*
     * free algorithm's memory, using the scratch-group Id for determining
     * shared scratch heap.
     */

    Log_print0(Diags_USER2,"[+2] DSKT2_freeAlg> Free instance memory");

    status = _DSKT2_freeInstanceMemory(scratchMutexId, memTab, numRecs,
            extHeap);

    /* done with stashed away memTab, free it */
    Memory_free(_DSKT2_heap, memTab, memTabSize);

    Log_print0(Diags_EXIT, "[+X] DSKT2_freeAlg> Exit");

    _DSKT2_algReferenceCount--;

    return (status);
}

#ifdef xdc_target__isaCompatible_64P
static inline Uns MC_ADDR_GLOB_TO_LOC(Uns glob_addr)
{  

   Uns tmp = glob_addr;

   //if ((tmp & 0xFF000000) == 0) {
   if ((tmp >> 24) == ((1 << 4)|DNUM)) {

      return (tmp & 0x00FFFFFF);

   } 

   else {

     return glob_addr;

   }
}
#endif
/*
 *  @(#) ti.sdo.fc.dskt2; 1, 0, 4,381; 11-2-2010 18:25:23; /db/atree/library/trees/fc/fc-n22x/src/ xlibrary

 */

