/* 
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/**
 *  @file       ti/sdo/fc/ecpy/ecpy.h
 *
 *  @brief      ECPY Definitions (C64P) - Functional DMA Library for use
 *              with the EDMA3 DMA peripheral.
 *              Provides a comprehensive list of DMA operations an algorithm
 *              can perform on logical DMA channels it acquires through the
 *              IRES_EDMA3CHAN protocol.
 *
 *  WARNING:    THIS IS NOT THE FINAL REVISION OF ECPY INTERFACE. IT IS SUBJECT
 *              TO CHANGES PER REVIEW FEEDBACK. ADDITIONALLY, CURRENTLY IT IS 
 *              AN API DEFINITION ONLY, IMPLEMENTATION AND COLLATERAL SUPPORT
 *              WILL BE ADDED IN FUTURE FC RELEASES.
 */

/**
 *  @defgroup   ti_sdo_fc_ecpy_ECPY ECPY (C64P)
 *
 *              The ECPY module provides a comprehensive list of DMA
 *              operations an algorithm can perform on logical DMA channels
 *              it acquires through the IRES_EDMA3CHAN protocol. Example of ECPY
 *              operations include channel configuration, DMA transfer
 *              scheduling, and DMA transfer synchronization.
 *
 *  @remark     For additional information, please refer to the
 *              "Using DMA with Framework Components for 'C64x+" document.
 */
#ifndef ti_sdo_fc_ecpy_ECPY_
#define ti_sdo_fc_ecpy_ECPY_


#ifdef __cplusplus
extern "C" {
#endif

/** @ingroup    ti_sdo_fc_ecpy_ECPY */
/*@{*/

#define ti_sdo_fc_ires_NOPROTOCOLREV 
#include <ti/sdo/fc/ires/edma3chan/ires_edma3Chan.h>

#define ECPY_GTNAME "ti.sdo.fc.ecpy"

/**
 * @brief       Name to pass to Diags_setMask() to enable logging for ECPY
 *              functions. For example,
 *                  Diags_setMask(ECPY_MODNAME"+EX1234567");
 *              turns on all Log statements in this module.
 *              Diags_setMask() must be called after initialization to take
 *              effect.
 */
#define ECPY_MODNAME "ti.sdo.fc.ecpy"


/**
 *  @brief  ECPY 16-bit param field structure. These values are passed to
 *          ECPY_configure16() to indicate the field of the
 *          ECPY_Params structure to be changed.
 */
typedef enum ECPY_ParamField16b {

    ECPY_PARAMFIELD_ELEMENTSIZE =  8,
    ECPY_PARAMFIELD_NUMELEMENTS = 10 ,
    ECPY_PARAMFIELD_ELEMENTINDEX_SRC = 16,
    ECPY_PARAMFIELD_ELEMENTINDEX_DST = 18,
    ECPY_PARAMFIELD_FRAMEINDEX_SRC =  24,
    ECPY_PARAMFIELD_FRAMEINDEX_DST =  26,
    ECPY_PARAMFIELD_NUMFRAMES = 28

} ECPY_ParamField16b;

/**
 *  @brief      ECPY 32-bit param field structure. These values are passed to
 *          ECPY_configure32() to indicate the field of the
 *          ECPY_Params structure to be changed.
 */
typedef enum ECPY_ParamField32b {

    ECPY_PARAMFIELD_SRCADDR = 4,
    ECPY_PARAMFIELD_DSTADDR = 12,
    ECPY_PARAMFIELD_ELEMENTINDEXES = 16,
    ECPY_PARAMFIELD_FRAMEINDEXES = 24

} ECPY_ParamField32b;

/**
 *  @brief      ECPY DMA Transfer Types:
 *
 *  Note that all transfers can be specified using the ECPY_2D2D transfer
 *  type, and providing element and frame indices for both source and
 *  destination. However, the other transfer types are provided to simplify
 *  configuration when applicable. For example, when using the ECPY_1D2D
 *  transfer type, you do not need to set the source element and frame
 *  indices in the ECPY_Params structure.
 */
typedef enum ECPY_TransferType {

    /**
     *  Only one element is copied from source to destination.
     *  The size of the element is specified in the elementSize field of
     *  ECPY_Params
     */
    ECPY_1D1D,

    /**
     *  After each element is copied to destination, the source
     *  and destination of the next element to copy is updated
     *  as follows:
     *  - src = src + element size
     *  - dst = dst + destination element index
     *
     *  After an entire frame is copied (the frame size specified by the
     *  numElements field of ECPY_Params), the source and destination of
     *  the next frame to copy are given by:
     *  - src = src + element size
     *  - dst = start of frame + destination frame index
     */
    ECPY_1D2D,

    /**
     *  This is similar to ECPY_1D2D, except that source and
     *  destination are updated after an element is copied,
     *  as follows:
     *  - src = src + source element index
     *  - dst = dst + element size
     *
     *  After an entire frame is copied by:
     *  - src = start of frame + source frame index
     *  - dst = dst + element size
     */
    ECPY_2D1D,

    /**
     *  This transfer type combines ECPY_1D2D and ECPY_2D1D,
     *  so that source and destination are updated after an
     *  element is copied by:
     *  - src = src + source element index
     *  - dst = dst + destination element index
     *
     *  After a frame is copied by:
     *  - src = start of frame + source frame index
     *  - dst = start of frame + destination frame index
     */
    ECPY_2D2D
} ECPY_TransferType;

/**
 *  @brief  ECPY_Obj defines the structure of the dma channel handle that
 *          is used to manage state about the EDMA3 channel.
 */
typedef struct  ECPY_Obj {

    IRES_EDMA3CHAN2_Obj *iresHandle;

    unsigned int iprRegisterAddr; /* Interrupt pending register */
    unsigned int iprBitMask;      /* Bit mask contining '1' corr to TCC */
    unsigned int esrRegisterAddr; /* Event Set Register Addr corr to channel */
    unsigned int esrBitMask;      /* Bit mask contining '1' corr to channel */
    short transferMode;           /* EDMA or QDMA mode transfers */    
    short transferPending;        
    short finalTransferNo;        /* TransferNo of the last transfer in a linked
                                   * transfer 
                                   */
    IRES_EDMA3CHAN_PaRamStruct *firstPARAMShadow;

} ECPY_Obj;

/**
 *  @brief      Handle to "logical" ECPY Channel
 */
typedef struct ECPY_Obj *ECPY_Handle;


/**
 *  @brief  DMA transfer specific parameters.  Defines the configuration
 *          of a logical channel.
 */
typedef struct ECPY_Params {

    /** 1D1D, 1D2D, 2D1D or 2D2D */
    ECPY_TransferType transferType;

    /** Source Address of the DMA transfer */
    Void *         srcAddr;

    /** Destination Address of the DMA transfer */
    Void *         dstAddr;

    /** Number of consecutive bytes in each 1D transfer vector (ACNT)   */
    unsigned short elementSize;

    /** Number of 1D vectors in 2D transfers (BCNT) */
    unsigned short numElements;

    /** Number of 2D frames in 3D transfers (CCNT) */
    unsigned short numFrames ;

    /**
     *  Offset in number of bytes from beginning of each 1D vector to the
     *  beginning of the next 1D vector.  (SBIDX)
     */
    short srcElementIndex;

    /**
     *  Offset in number of bytes from beginning of each 1D vector to the
     *  beginning of the next 1D vector.  (DBIDX)
     */
    short dstElementIndex;

    /**
     *  Offset in number of bytes from beginning of 1D vector of current
     *  (source) frame to the beginning of next frame's first 1D vector.
     *  (SCIDX): signed value between -32768 and 32767.
     */
    short srcFrameIndex;

    /**
     *  Offset in number of bytes from beginning of 1D vector of current
     *  (destination) frame to the beginning of next frame's first 1D vector.
     *  (DCIDX): signed value between -32768 and 32767.
     */
    short dstFrameIndex;

} ECPY_Params;


/**
 *  @brief      Configure all DMA transfer settings for the logical channel.
 *
 *  @param[in]  handle      ECPY channel handle.
 *  @param[in]  params      DMA transfer specific parameters used to
 *                          configure this logical DMA channel.
 *  @param[in]  transferNo  Indicates the individual transfer to be
 *                          configured using the passed 'params'
 *
 *  @pre        channel must be in active state
 *  @pre        valid IDMA_Obj handle
 *  @pre        handle->protocol == &ECPY_PROTOCOL
 *  @pre        0 < transferNo <= originally requested number of transfers
 */
Void ECPY_configure(ECPY_Handle handle, ECPY_Params * restrict params,
    short transferNo);
Void ECPY_directConfigure(ECPY_Handle handle, ECPY_Params * restrict params,
    short transferNo);


/**
 *  @brief      Modify the 16-bit DMA transfer parameter, indicated by the
 *              parameter field id, fieldId, of the current channel settings.
 *
 *  @param[in]  handle      IRES_EDMA3CHAN channel handle.
 *  @param[in]  fieldId     Indicates which of the parameters is to be
 *                          modified.
 *  @param[in]  value       New value of the parameter to be modified.
 *  @param[in]  transferNo  Indicates which transfer the parameters correspond
 *                          to (same value that would be passed to
 *                          ECPY_configure()).
 *
 *  @pre        Algorithm instance must be in "active" state using IALG
 *              interface.
 *
 *  @pre        Channel must be in "active" state.
 *
 *  @remarks    Once a channel has been configured once with ECPY_configure(),
 *              ECPY_configure16() can be used to update any of the
 *              16-bit parameter fields, for example, the number of elements
 *              in a frame to be transferred.
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_CONFIGURE16 or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
*/
#if defined(ECPY_CONFIGURE16) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_configure16.h>
#else
Void ECPY_configure16(ECPY_Handle handle, ECPY_ParamField16b fieldId, 
        unsigned short value, short transferNo);
Void ECPY_directConfigure16(ECPY_Handle handle, ECPY_ParamField16b fieldId, 
        unsigned short value, short transferNo);
#endif

/**
 *  @brief      Modify the 32-bit DMA transfer parameter, indicated by the
 *              parameter field id, fieldId, of the current channel settings.
 *
 *  @param[in]  handle      ECPY  channel handle.
 *  @param[in]  fieldId     Indicates which of the parameters is to be
 *                          modified.
 *  @param[in]  value       New value of the parameter to be modified.
 *  @param[in]  transferNo  Indicates which transfer the parameters correspond
 *                          to (same value that would be passed to
 *                          ECPY_configure()).
 *
 *  @pre        @c handle must be a valid channel handle.
 *
 *  @pre        Algorithm instance must be in "active" state using IALG
 *              interface.
 *
 *  @pre        Channel must be in 'active' state.
 *
 *  @remarks    Once a channel has been configured once with ECPY_configure(),
 *              ECPY_configure32() can be used to update any of the
 *              32-bit parameter fields, for example, the source address of
 *              the data to be transferred.
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_CONFIGURE32_INLINE or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
 *
*/
#if defined(ECPY_CONFIGURE32_INLINE) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_configure32.h>
#else
Void ECPY_configure32(ECPY_Handle handle, ECPY_ParamField32b fieldId, 
        unsigned int value, short transferNo);
Void ECPY_directConfigure32(ECPY_Handle handle, ECPY_ParamField32b fieldId, 
        unsigned int value, short transferNo);
#endif

/**
 *  @brief      Indicate that a given transfer will be the last in a sequence
 *              of linked transfers.
 *
 *              This API can be used to dynamically change the number of
 *              transfers in a linked transfers. 
 *
 *  @param[in]  handle      ECPY channel handle.
 *  @param[in]  transferNo  Indicates which transfer that will be the last
 *                          in the set of linked transfers. (This is the same
 *                          value that was passed to ECPY_configure()).
 *
 *  @remarks    This API can be used if a channel was created to
 *              transfer numTransfers linked transfers, but at
 *              some point, it may be that fewer transfers than
 *              numTransfers should be started.
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_SETFINAL_INLINE or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
 *
 *  @pre        Valid channel handle.
 *  @pre        Channel must be in 'active' state.
 *
 */
#if defined(ECPY_SETFINAL_INLINE) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_setfinal.h>
#else
Void ECPY_setFinal(ECPY_Handle handle, short transferNo);
Void ECPY_directSetFinal(ECPY_Handle handle, short transferNo);
#endif

/**
 *  @brief      Indicate the memory space that DMA transfers submitted to
 *              the channel associated with the handle will modify.
 *              This information can be used for channel to TC Queue mapping
 *              for avoiding known Silicon GEM lockup issues or for 
 *              performance improvements.
 *
 *  @param[in]  handle  ECPY channel handle.
 *  @param[in]  mode    DMA destination description of type 
 *                      IRES_EDMA3CHAN_DmaDestType.
 */
Void ECPY_setDestinationMemoryMode(ECPY_Handle handle, short mode);

/**
 *  @brief      Submit a single or linked transfer using the most recently
 *              configured transfer parameter settings.
 *
 *  @param[in]  handle  ECPY channel handle.
 *
 *  @pre        Valid channel handle.
 *  @pre        Channel must be in 'active' state.
 *  @pre        Channel must be in 'configured' state.
 *
 *  @post       handle->transferPending = TRUE
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_START_INLINE or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
 *
 */
#if defined(ECPY_START_INLINE) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_start.h>
#else
Void ECPY_start(ECPY_Handle handle);
Void ECPY_directStartEdma(ECPY_Handle handle);
Void ECPY_directStartQdma(ECPY_Handle handle);
#endif


/**
 *  @brief      Wait for all data transfers on a logical channel to complete.
 *
 *
 *  @param[in]  handle  ECPY channel handle.
 *
 *  @pre        Valid channel handle
 *  @pre        Channel must be in 'active' state.
 *  @pre        ECPY_wait can be called back to back or without having submitted
 *              transfers to the channel. For the case there is no pending 
 *              transfer on the channel it will return immediately.
 *
 *  @post       ECPY_wait will reset channel handle's transferPending flag to 0
 *
 *  @remarks    This function will not return until all the previously submitted
 *              EDMA transfers on the channel have completed.
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_WAIT_INLINE or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
 */
#if defined(ECPY_WAIT_INLINE) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_wait.h>
#else
Void ECPY_wait(ECPY_Handle handle);
#endif

/**
 *  @brief      Wait for all EDMA data transfers submitted to the logical 
 *              channel via ECPY_directStartEdma to complete.
 *
 *
 *  @param[in]  handle  ECPY channel handle.
 *
 *  @pre        Valid channel handle
 *  @pre        Channel must be in 'active' state.
 *  @pre        ECPY_directWait requires that a transfer has been submitted 
 *              to the channel using ECPY_directStartEdma API. Otherwise it will
 *              block indefinitely.
 *  @pre        ECPY_wait can not be called back to back or without having 
 *              started EDMA transfers on the channel. 
 *
 *  @remarks    ECPY_directWait will not maintain channel handle's 
 *              transferPending state.
 *
 *  @remarks    This function will not return until all the data
 *              transfers on the given channel have completed.
 *
 *  @remarks    Both an inlined version and a non-inlined version of this
 *              function are supported. To use the inlined version, add the
 *              line \#define ECPY_DIRECTWAIT_INLINE or
 *              \#define ECPY_INLINE_ALL to the source file where this
 *              function is used. The \#define must go above the line,
 *              \#include <ecpy.h>. By default, this function will not be
 *              inlined. Another alternative is to define the macro in the
 *              build option of the file or of the project file.
 */
#if defined(ECPY_DIRECTWAIT_INLINE) || defined(ECPY_INLINE_ALL)
#include <ti/sdo/fc/ecpy/ecpy_directwait.h>
#else
Void ECPY_directWait(ECPY_Handle handle);
#endif

/**
 *  @brief      Activate the given channel.
 *
 *  @param[in]  handle  ECPY channel handle.
 *
 *  @pre        @c handle must be a valid channel handle.
 *
 *  @pre        Algorithm instance must be in "active" state using IALG
 *              interface.
 *
 *  @pre        @c handle must be a valid handle obtained by ECPY_createHandle.
 *
 *  @post       Channel is in 'active' state, any ECPY APIs can be called
 *              using this handle.
 */
Void ECPY_activate(ECPY_Handle handle);

/**
 *  @brief      Create an ECPY Handle using the supplied IRES_EDMA3CHAN handle
 *
 *  @param[in]  handle      IRES_EDMA3CHAN channel handle.
 *  @param[in]  alg         IALG handle.
 *
 *  @pre        @c handle must be a valid handle.
 *
 */
ECPY_Handle ECPY_createHandle(IRES_EDMA3CHAN2_Handle handle, IALG_Handle alg);

/**
 *  @brief      Delete ECPY_Handle
 *
 *  @param[in]  handle      ECPY channel handle.
 *
 *  @pre        @c handle must be a valid handle obtained by ECPY_createHandle.
 *
 */
Void ECPY_deleteHandle(ECPY_Handle handle);

/**
 *  @brief      Deactivates the given channel.
 *
 *  @param[in]  handle    ECPY  channel handle.
 *
 *  @pre        @c handle must be a valid channel handle.
 *
 *  @pre        Channel must be in 'active' state.
 *
 *  @pre        Algorithm instance must be in 'active' state using IALG
 *              interface
 *
 *  @post       Channel is in 'deactivated' state.
 */
Void ECPY_deactivate(ECPY_Handle handle);


/**
 *  @brief              Initialize the ECPY module.  Called by framework.
 */
Void ECPY_init(Void);

/**
 *  @brief              Finalization of the ECPY module.
 */
Void ECPY_exit(Void);

/**
 *  @brief      Memory allocation function signature for module configuration.
 */
typedef Bool (ECPY_PersistentAllocFxn)(IALG_MemRec *memTab, Int numRecs);

/**
 *  @brief      Memory deallocation function signature for module configuration.
 */
typedef Void (ECPY_PersistentFreeFxn)(IALG_MemRec *memTab, Int numRecs);


/**
 *  @brief      The module configuration structure for ECPY.
 *              It is set at design time by the system integrator to register
 *              customize memory allocation/free implementations.
 *
 *  @remarks    Typically, system integrators configure these settings using
 *              XDC config scripts.  System integrators that use that method
 *              of configuration can ignore this structure definition.  It's
 *              auto-generated during the config step by the XDC Tools.
 */
typedef struct ECPY_CfgParams {

    /**
     *  @brief      Memory allocation function for all ECPY memory requirements
     */
    ECPY_PersistentAllocFxn * allocFxn;

    /**
     *  @brief      Memory free function for all ECPY memory requirements
     */
    ECPY_PersistentFreeFxn * freeFxn;

} ECPY_CfgParams;

/**
 *  @brief      Initialization parameters for the ECPY module
 */
extern __FAR__ ECPY_CfgParams ECPY_CFG_PARAMS;

/*@}*/

#ifdef __cplusplus
}
#endif /* extern "C" */


#endif /* ti_sdo_fc_ecpy_ECPY_ */

/*
 *  @(#) ti.sdo.fc.ecpy; 1, 0, 0,54; 11-2-2010 18:25:29; /db/atree/library/trees/fc/fc-n22x/src/ xlibrary

 */

