/* 
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/**
 *  @file       ti/sdo/fc/ecpy/ecpy_impl.c
 *
 *  @brief      ECPY Definitions (C64P) - EDMA functional library  
 *              Provides a comprehensive list of DMA operations an algorithm
 *              can perform on logical DMA channels it acquires through the
 *              IRES_EDMA3CHAN protocol.
 *
 *  WARNING:    THIS IS NOT THE FINAL REVISION OF ECPY INTERFACE. IT IS SUBJECT
 *              TO CHANGES PER REVIEW FEEDBACK. ADDITIONALLY, CURRENTLY IT IS 
 *              AN API DEFINITION ONLY, IMPLEMENTATION AND COLLATERAL SUPPORT
 *              WILL BE ADDED IN FUTURE FC RELEASES.
 */

/**
 *  @defgroup   ti_sdo_fc_ecpy_ECPY ECPY (C64P)
 *
 *              The ECPY module provides a comprehensive list of DMA
 *              operations an algorithm can perform on logical DMA channels
 *              it acquires through the IRES_EDMA3CHAN protocol. Example of ECPY
 *              operations include channel configuration, DMA transfer
 *              scheduling, and DMA transfer synchronization.
 *
 *  @remark     For additional information, please refer to the
 *              "Using DMA with Framework Components for 'C64x+" document.
 */

#ifdef __cplusplus
extern "C" {
#endif

#define Registry_CURDESC ti_sdo_fc_ecpy_desc

/** @ingroup    ti_sdo_fc_ecpy_ECPY */
/*@{*/
#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Registry.h>
#include "ecpy.h"
#include "ecpy_util.h"

static Void *myAlloc(size_t size, size_t align);
static Void myFree(Ptr addr, size_t size);
extern Registry_Desc ti_sdo_fc_ecpy_desc;

/*
 *  ======== myAlloc ========
 *  Wrapper around ECPY_CFG_PARAMS.allocFxn.
 */
static Void *myAlloc(size_t size, size_t align)
{
    IALG_MemRec memTab;
    Void       *addr = NULL;

    if (NULL == ECPY_CFG_PARAMS.allocFxn) {
        return (NULL);
    } 

    memTab.size = size;
    memTab.alignment = align;
    memTab.space = IALG_EXTERNAL;
    memTab.attrs = IALG_PERSIST;

    if (ECPY_CFG_PARAMS.allocFxn(&memTab, 1)) {
        addr = memTab.base;
    }

    return (addr);
}

/*
 *  ======== myFree ========
 *  Wrapper around ECPY_CFG_PARAMS.freeFxn.
 */
static Void myFree(Ptr addr, size_t size)
{
    IALG_MemRec memTab;

    memTab.base = addr;
    memTab.size = size;
    memTab.alignment = 0;         /* Alignment shouldn't be needed here */
    memTab.space = IALG_EXTERNAL;
    memTab.attrs = IALG_PERSIST;

    ECPY_CFG_PARAMS.freeFxn(&memTab, 1);
}

/*
 *  ======== ECPY_getPaRamLinkIndex ============
 *  Returns the PARAM ID of the Physical PaRAM Regs associated with the
 *  logical channel handle's transfer for configuring the LINK field in linked
 *  transfers.
 */
static __inline
Uns ECPY_getPaRamLinkIndex(IRES_EDMA3CHAN2_Handle handle, short transferNo)
{
    return ((Uns)(handle->assignedPaRamAddresses[transferNo]) & 
            0xFFFF);                 
}

static IRES_EDMA3CHAN_EDMA3RegisterLayer *_EDMA3_globalRegs = NULL;

static IRES_EDMA3CHAN_Properties  _EDMA3CHAN_staticProperties;

static void _ECPY_initEDMA3CHANStaticProperties(IRES_EDMA3CHAN2_Handle handle)
{
        if (_EDMA3_globalRegs == NULL) {

            handle->ires.getStaticProperties((struct IRES_Obj *)handle,  
                    (IRES_Properties *)&_EDMA3CHAN_staticProperties);
            _EDMA3_globalRegs = _EDMA3CHAN_staticProperties.globalRegs;
    }
}
        

/* common PARAM update portion of _configure and _directConfigure functions */
static __inline Void
_updatePaRam(ECPY_Handle handle, IRES_EDMA3CHAN_PaRamStruct *paRam, 
        ECPY_Params * restrict params, short transferNo)
{
    unsigned short   tcc;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    /* #1 - Source address */
    paRam->src = (unsigned int)params->srcAddr;

    /* #2 */
    paRam->acnt = params->elementSize;
    paRam->bcnt = 1; /* default for 1D1D */

    /* #3 - Destination address */
    paRam->dst = (unsigned int)params->dstAddr;

    /* #5 - Set bcnt reload value to bcnt, set to no link */
    paRam->bCntrld = params->numElements;

    if (params->transferType != ECPY_1D1D) {

        /* Field #2: update with 2nd dimension */
        paRam->bcnt = params->numElements;

        /* Field #3: set Element Indexes for 2D2D */
        paRam->srcElementIndex = params->srcElementIndex;
        paRam->dstElementIndex = params->dstElementIndex;

        if (params->transferType == ECPY_1D2D) {
            Log_print0(Diags_USER4, "[+4] ECPY_configure> 1D2D transfer");
            paRam->srcElementIndex = params->elementSize;

        }
        else if (params->transferType == ECPY_2D1D) {
            Log_print0(Diags_USER4, "[+4] ECPY_configure> 2D1D transfer");
            paRam->dstElementIndex = params->elementSize;
        
        } else {

            Log_print0(Diags_USER4, "[+4] ECPY_configure> 2D2D transfer");
        }

        paRam->srcFrameIndex = params->srcFrameIndex;
        paRam->dstFrameIndex = params->dstFrameIndex;

    } else {

        Log_print0(Diags_USER4, "[+4] ECPY_configure> 1D1D transfer");
    }


    if (0 == params->numFrames) {
        paRam->ccnt = 1;
    }
    else {
        paRam->ccnt = params->numFrames;
    }

    Log_print4(Diags_USER4, "[+4] ECPY_configure> src 0x%x, "
            "dst 0x%x, acnt 0x%x, bcnt 0x%x ",
            (IArg)(paRam->src), (IArg)(paRam->dst),
            (IArg)(paRam->acnt), (IArg)(paRam->bcnt));

    Log_print3(Diags_USER4, "[+4] ECPY_configure> bcnt reload"
            " 0x%x, srcIndex 0x%x, dstIndex 0x%x",
            (IArg)(paRam->bCntrld), (IArg)(paRam->srcElementIndex),
            (IArg)(paRam->dstElementIndex));

    /* Pick the tcc from the handle */
    tcc = eHandle->assignedTccIndices[0];
    Log_print1(Diags_USER2, "[+2] ECPY_configure> Tcc %d", (IArg)tcc);
 
    /*
     * If configuring handle with linked transfers then
     * adjust the 'link' and 'tcc' synchronization, etc., in opt.
     */
    if (transferNo < (eHandle->assignedNumPaRams)) {

        /* #5 */
        /*
         *  Setup this param's link field to point to the NEXT transfer when
         *  setting up linked transfers we assign PARAM addresses in
         *  descending order, so that the last entry is the first transfer,
         *  etc., handle's paRamAddr is pointing to the 'last' PARAM entry
         *  assigned to the channel handle (along with the preceding
         *  "numTransfer" PARAM entries.
         */
        paRam->link = (unsigned short)
                ECPY_getPaRamLinkIndex(eHandle, transferNo);

        /* 
         * 3D transfers require that intermediate transfer chaining (ITCCHEN)
         * is enabled and STATIC bit is 0.
         * Linked transfer require  transfer chaining (TCCHEN) is enabled.
         */
        paRam->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                     ECPY_PARAMOPTS_ITCCHEN | /* intermediate tr chaining ON*/
                     ECPY_PARAMOPTS_TCCHEN | 
                     ECPY_PARAMOPTS_ABSYNC;
                     
        Log_print2(Diags_USER4, "[+4] ECPY_configure> Link this"
                " transfer to 0x%x Opt is 0x%x",
                (IArg)(paRam->link), (IArg)(paRam->opt));
    }
    else {
        /* Single transfer or the last one of the Linked transfers */
        /* #5 - Set no link for final PaRam */
        paRam->link = 0xffff;

        /* #0 */
        /* Final transfer requires  transfer chaining (TCCHEN) is disabled. */
        paRam->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                     ECPY_PARAMOPTS_ABSYNC       |
                     ECPY_PARAMOPTS_ITCCHEN | /* intermediate tr chaining ON*/
                     ECPY_PARAMOPTS_TCC_INTEN;

        Log_print2(Diags_USER4, "[+4] ECPY_configure> Mark as Final "
                " transfer setting Link to 0x%x Opt is 0x%x",
                (IArg)(paRam->link), (IArg)(paRam->opt));
    }
}
/**
 *  @brief      Configure all DMA transfer settings for the logical channel.
 *
 *  @param[in]  handle      IRES_EDMA3CHAN channel handle.
 *  @param[in]  params      DMA transfer specific parameters used to
 *                          configure this logical DMA channel.
 *  @param[in]  transferNo  Indicates the individual transfer to be
 *                          configured using the passed 'params'
 *
 *  @pre        channel must be in active state
 *  @pre        valid  handle
 *  @pre        0 < transferNo <= originally requested number of transfers
 */
Void ECPY_configure(ECPY_Handle handle, ECPY_Params * restrict params,
    short transferNo)
{
    IRES_EDMA3CHAN_PaRamStruct * restrict paRam;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    Log_print2(Diags_ENTRY, "[+E] ECPY_configure> Enter "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)transferNo);
    
#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);

    if ((transferNo == 0) || (transferNo > eHandle->assignedNumPaRams)) {

        Log_print2(Diags_USER4, "[+E] ECPY_configure> ERROR! (transferNo=%d) "
            " should be non-Zero and less than assignedNumPaRams:%d", 
            (IArg)transferNo, (IArg)eHandle->assignedNumPaRams);
    }
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
    Assert_isTrue((transferNo > 0), (Assert_Id)NULL);

#endif /* _DEBUG_ defined only for debug profile */

    
    /*
     * Always update shadow params of the 1st transfer when using Linked
     * transfers. The physical PaRam of the 1st trasfer in a Linked xfer
     * gets overwritten copied  by the Linked transfers PaRAMs, so it needs
     * to be saved and refreshed before 'start'.
     */
    if ( (eHandle->assignedNumPaRams > 1) && (transferNo == 1) ) {
        /* 1st transfer updates the shadow PaRAM */
        paRam = (IRES_EDMA3CHAN_PaRamStruct *)handle->firstPARAMShadow;

    } else {
        /* modify the assigned Physical PaRAM directly */
        paRam = (IRES_EDMA3CHAN_PaRamStruct *)
                eHandle->assignedPaRamAddresses[transferNo-1];
    }

    _updatePaRam(handle, paRam, params, transferNo);

    Log_print2(Diags_EXIT, "[+E] ECPY_configure> Exit "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)transferNo);
}

/**
 *  @brief      Configure all DMA transfer settings for the logical channel.
 *
 *  @param[in]  handle      IRES_EDMA3CHAN channel handle.
 *  @param[in]  params      DMA transfer specific parameters used to
 *                          configure this logical DMA channel.
 *  @param[in]  transferNo  Indicates the individual transfer to be
 *                          configured using the passed 'params'
 *
 *  @pre        channel must be in active state
 *  @pre        valid  handle
 *  @pre        0 < transferNo <= originally requested number of transfers
 */
Void ECPY_directConfigure(ECPY_Handle handle, ECPY_Params * restrict params,
    short transferNo)
{
    IRES_EDMA3CHAN_PaRamStruct * restrict paRam;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    Log_print2(Diags_ENTRY, "[+E] ECPY_directConfigure> Enter "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)transferNo);
    
#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);

    if ((transferNo == 0) || (transferNo > eHandle->assignedNumPaRams)) {

        Log_print2(Diags_USER4, "[+E] ECPY_configure> ERROR! (transferNo=%d) "
            " should be non-Zero and less than assignedNumPaRams:%d", 
            (IArg)transferNo, (IArg)eHandle->assignedNumPaRams);
    }
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
    Assert_isTrue(transferNo > 0, (Assert_Id)NULL);

#endif /* _DEBUG_ defined only for debug profile */


    /*
     * Fast version of configure, directly updating Physical EDMA3 PARAM space.
     * Does not update shadow params of the 1st transfer when using Linked
     * transfers. The physical PaRam of the 1st trasfer in a Linked xfer
     * gets overwritten copied  by the Linked transfers PaRAMs, so it needs
     * to be saved and refreshed before 'start'.
     */
     /* modify the assigned Physical PaRAM directly */
    paRam = (IRES_EDMA3CHAN_PaRamStruct *)
            eHandle->assignedPaRamAddresses[transferNo-1];

    _updatePaRam(handle, paRam, params, transferNo);

    Log_print2(Diags_EXIT, "[+E] ECPY_directConfigure> Exit "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)transferNo);
}

/*
 *  ======== ECPY_directConfigure16 ========
 */
Void ECPY_directConfigure16(ECPY_Handle restrict handle,
    ECPY_ParamField16b fieldId, unsigned short value, short transferNo)
{
    IRES_EDMA3CHAN_PaRamStruct * restrict paRam;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    Log_print4(Diags_ENTRY, "[+E] ECPY_directConfigure16b> Enter"
            " (handle=0x%x, fieldId=0x%x, value=0x%x, trasnferNo=%d",
            (IArg)handle, (IArg)fieldId, (IArg)value, (IArg)transferNo);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
    Assert_isTrue(transferNo > 0, (Assert_Id)NULL);
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    paRam = (IRES_EDMA3CHAN_PaRamStruct *)
            eHandle->assignedPaRamAddresses[transferNo-1];

    *((unsigned short *)((Uns)paRam + fieldId)) = value;

    Log_print0(Diags_EXIT, "[+E] ECPY_directConfigure16> Exit ");
}

/*
 *  ======== ECPY_directConfigure32 ========
 */
Void ECPY_directConfigure32(ECPY_Handle restrict handle,
    ECPY_ParamField32b fieldId, unsigned int value, short transferNo)
{
    IRES_EDMA3CHAN_PaRamStruct * restrict paRam;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
        (handle->iresHandle);

    Log_print4(Diags_ENTRY, "[+E] ECPY_directConfigure32b> Enter"
            " (handle=0x%x, fieldId=0x%x, value=0x%x, trasnferNo=%d",
            (IArg)handle, (IArg)fieldId, (IArg)value, (IArg)transferNo);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
    Assert_isTrue(transferNo > 0, (Assert_Id)NULL);
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    paRam = (IRES_EDMA3CHAN_PaRamStruct *)eHandle->
            assignedPaRamAddresses[transferNo-1];

    *((unsigned int *)((unsigned int)paRam + fieldId)) = value;

    Log_print0(Diags_EXIT, "[+E] ECPY_directConfigure32> Exit ");
}

/**
 *  @brief      Indicate that a given transfer will be the last in a sequence
 *              of linked transfers.
 *
 *              This API can be used to dynamically change the number of
 *              transfers in a linked transfers. 
 *
 *  @param[in]  handle      ECPY channel handle.
 *  @param[in]  transferNo  Indicates which transfer that will be the last
 *                          in the set of linked transfers. (This is the same
 *                          value that was passed to ECPY_configure()).
 *
 *  @remarks    This API can be used if a channel was created to
 *              transfer numTransfers linked transfers, but at
 *              some point, it may be that fewer transfers than
 *              numTransfers should be started.
 *
 *  @remarks    This API configures both 'direct' params as well as shadow
 *              paRAMs.
 *
 *  @pre        Valid channel handle.
 *  @pre        Channel must be in 'active' state.
 *
 *  @post       ECPY Handle's state is updated with 'current' last transferNo.
 *  @post       If current 'final' is > previous 'final' transfer, fix the 
 *              link field of the previous 'final' transfer to link to the 
 *              next transfer!
 */
Void ECPY_setFinal(ECPY_Handle handle, short transferNo)
{
    unsigned short tcc;
    IRES_EDMA3CHAN_PaRamStruct *paRamThis;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    short numTotalPaRams = eHandle->assignedNumPaRams; 
    short prevFinal = handle->finalTransferNo;

    Log_print2(Diags_ENTRY, "[+E] ECPY_setFinal> Enter "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)(transferNo ));

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
    Assert_isTrue(transferNo > 0, (Assert_Id)NULL);
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    /* Single transfer or the last one of the Linked transfers */
    tcc = eHandle->assignedTccIndices[0];
 
    if (numTotalPaRams == 1) {
        return;
    }

    handle->finalTransferNo = transferNo; /* this xfer becomes the final */

    /*
     *  Reset the link id of the previous transfer that was configured as
     *  final, unless it is the last PaRam allocated to the channel.
     */
    if ( (prevFinal < numTotalPaRams) && (prevFinal != transferNo)) {
     
        IRES_EDMA3CHAN_PaRamStruct *paRamPrevFinal;

        /* If prevFinal was 1st xfer, update the shadow PaRAMs . */
        if (prevFinal == 1) {
            paRamPrevFinal = handle->firstPARAMShadow;

        } else {
            /* modify the assigned Physical PaRAM directly */
            paRamPrevFinal = (IRES_EDMA3CHAN_PaRamStruct *)
                    eHandle->assignedPaRamAddresses[prevFinal-1];
        }
                
        paRamPrevFinal->link = (unsigned short)
                ECPY_getPaRamLinkIndex(eHandle, prevFinal);

        paRamPrevFinal->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                              ECPY_PARAMOPTS_TCCHEN | 
                              ECPY_PARAMOPTS_ITCCHEN | 
                              ECPY_PARAMOPTS_ABSYNC;
    }

    /* If setting 1st xfer as Final, update the shadow PaRAMs . */
    if (transferNo == 1) {
        paRamThis = handle->firstPARAMShadow;

    } else {
        /* modify the assigned Physical PaRAM directly */
        paRamThis = (IRES_EDMA3CHAN_PaRamStruct *)
                eHandle->assignedPaRamAddresses[transferNo-1];
    }

    /* Set no link for final PaRam */
    paRamThis->link = 0xffff;

    /* assign the TCC and enable TCC Interrupt bit for the final transfer */
    paRamThis->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                     ECPY_PARAMOPTS_ITCCHEN      | 
                     ECPY_PARAMOPTS_ABSYNC       |
                     ECPY_PARAMOPTS_TCC_INTEN    ;
    
    Log_print0(Diags_EXIT, "[+E] ECPY_setFinal> Exit ");
}
/*
 * ======== ECPY_directSetFinal ====================
 */
Void ECPY_directSetFinal(ECPY_Handle handle, short transferNo)
{
    unsigned short tcc;
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);
    IRES_EDMA3CHAN_PaRamStruct *paRamThis = (IRES_EDMA3CHAN_PaRamStruct *)
                eHandle->assignedPaRamAddresses[transferNo-1];
    short numTotalPaRams = eHandle->assignedNumPaRams; 
    short prevFinal = handle->finalTransferNo;
    

    Log_print2(Diags_ENTRY, "[+E] ECPY_directSetFinal> Enter "
            "(handle=0x%x, transferNo=%d)", (IArg)handle, (IArg)(transferNo ));

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
    Assert_isTrue(transferNo > 0, (Assert_Id)NULL);
    Assert_isTrue(transferNo <= eHandle->assignedNumPaRams, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    if (numTotalPaRams == 1) {
    
        Log_print0(Diags_EXIT, "[+E] ECPY_setFinal> Exit ");
        return;
    }

    handle->finalTransferNo = transferNo; /* this xfer becomes the final */

    /* Single transfer or the last one of the Linked transfers */
    tcc = eHandle->assignedTccIndices[0];

    /*
     *  Reset the link id of the previous transfer that was configured as
     *  final, unless it is the last PaRam allocated to the channel.
     */
    if ( (prevFinal < numTotalPaRams) && (prevFinal != transferNo)) {
     
        IRES_EDMA3CHAN_PaRamStruct *paRamPrevFinal;

        /* modify the assigned Physical PaRAM directly */
        paRamPrevFinal = (IRES_EDMA3CHAN_PaRamStruct *)
                eHandle->assignedPaRamAddresses[prevFinal-1];
                
        paRamPrevFinal->link = (unsigned short)
                ECPY_getPaRamLinkIndex(eHandle, prevFinal);

        paRamPrevFinal->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                              ECPY_PARAMOPTS_TCCHEN | 
                              ECPY_PARAMOPTS_ITCCHEN | 
                              ECPY_PARAMOPTS_ABSYNC;
    }

    /* Set no link for final PaRam */
    paRamThis->link = 0xffff;
 
    /* assign the TCC and enable TCC Interrupt bit for the final transfer */
    paRamThis->opt = ECPY_PARAMOPTS_TCCBITS(tcc) |   
                     ECPY_PARAMOPTS_ITCCHEN      | 
                     ECPY_PARAMOPTS_ABSYNC       |
                     ECPY_PARAMOPTS_TCC_INTEN    ;

    Log_print0(Diags_EXIT, "[+E] ECPY_directSetFinal> Exit ");
}

/**
 *  @brief      Submit a single or linked transfer using the most recently
 *              configured transfer parameter settings.
 *
 *  @param[in]  handle   channel handle.
 *
 *  @pre        Valid channel handle.
 *  @pre        Channel must be in 'active' state.
 *  @pre        Channel must be in 'configured' state.
 *
 */
Void ECPY_start(ECPY_Handle handle)
{
    IRES_EDMA3CHAN2_Handle eHandle = (IRES_EDMA3CHAN2_Handle)
            (handle->iresHandle);

    Log_print1(Diags_ENTRY, "[+E] ECPY_start> Enter (handle=0x%x)",
            (IArg)handle);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
    Assert_isTrue(handle->transferPending == FALSE, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    /* 
     * For performance uses the ESR Register pointer and channel bit mask
     * that is cached in the handle:
     */

    /* TODO: optimization. Add a 'linkedTransfer' field in ECPY handle
     * and use shadow based on this flag instead of assignedNumPaRams. 
     * reflecting 'setFinal' usage as well. 
     */
    if (eHandle->assignedNumPaRams > 1) {
        /* update the 1st transfer's PaRAMs from its shadowParams */
        IRES_EDMA3CHAN_PaRamStruct * restrict paRam;
        IRES_EDMA3CHAN_PaRamStruct * shadow_paRam;
        int i;
    
        paRam =(IRES_EDMA3CHAN_PaRamStruct *)
                eHandle->assignedPaRamAddresses[0];
        shadow_paRam = (IRES_EDMA3CHAN_PaRamStruct *)handle->firstPARAMShadow;

       /* Copy shadow into Phys PaRam registers (must use 32-bit writes) */
       for (i = 0; i < 8; i++) {
            *((Uns *)paRam + i) = *((Uns *)shadow_paRam + i);
       }
    }
    /* Set transferPending indicating a transfer has been submitted */
    handle->transferPending = 1;
    
    /* Manually trigger EDMA transfer by setting the ESR bit corresponding
     * to the EDMA channel 
     */
    *(volatile unsigned int *)handle->esrRegisterAddr = handle->esrBitMask;

    Log_print0(Diags_EXIT, "[+E] ECPY_start> Exit ");
}

/**
 *  @brief      Submit a single or linked transfer using the most recently
 *              modified EDMA3 Physical PaRAM transfer parameter settings.
 *
 *  @param[in]  handle   channel handle.
 *
 *  @pre        Valid channel handle.
 *  @pre        Channel must be in 'active' state.
 *  @pre        Channel must be in 'configured' state.
 *
 */
Void ECPY_directStartEdma(ECPY_Handle handle)
{

    Log_print1(Diags_ENTRY, "[+E] ECPY_directStartEdma> Enter (handle=0x%x)",
            (IArg)handle);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */
    
    /* 
     * Starts an EDMA manual triggered transfer using the most recent 
     * Physical EDMA PARAM values. If the previous transfer was a Linked
     * transfer, the 1st PARAM must be directConfigure()'ed before calling
     * this function. This functions does not use shadow PaRAMs to retain
     * logical channel PARAM settings.
     */ 
    /* Set transferPending indicating a transfer has been submitted */
    handle->transferPending = 1;

    /* Manually trigger EDMA transfer by setting the ESR bit corresponding
     * to the EDMA channel 
     */
    *(volatile unsigned int *)handle->esrRegisterAddr = handle->esrBitMask;

    Log_print0(Diags_EXIT, "[+E] ECPY_directStartEdma> Exit ");
}

/*
 * ECPY_createHandle(IRES_EDMA3CHAN2_Handle handle)
 */
ECPY_Handle ECPY_createHandle(IRES_EDMA3CHAN2_Handle handle, IALG_Handle alg)
{
    unsigned int channel;
    unsigned int tcc;
    ECPY_Handle ecpyHandle = NULL;

    Log_print2(Diags_ENTRY, "[+E] ECPY_createHandle> Enter "
            "(handle=0x%x, ialgHandle=%d)", (IArg)handle, (IArg)(alg ));

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */
    
    /* First time this API is called we initialize the EDMA3 Static Properties*/
    if (_EDMA3_globalRegs == NULL) {
            _ECPY_initEDMA3CHANStaticProperties(handle);
    }
    /* TODO: Currently nothing done for QDMA mode of operation. */

    /* 
     * ECPY no longer requires that the IRES_EDMA3CHAN has been obtained with
     * shadowParams  allocation. 
     * When channel already has shadow memory assigned  ECPY uses this shadow 
     * memory instead of internally allocating it, otherwise it allocates the
     * needed memory.
     *
     * ECPY_activate subsequently uses the allocated memory for the 
     * shadowParams to construct an ECPY_handle, as well as for
     * shadowing the 1st PaRAM of linked transfers.
     */
    ecpyHandle = (ECPY_Handle)handle->shadowPaRams;

    if (ecpyHandle == NULL) {

        /* 
         * Allocate memory for ECPY_Obj & shadow for 1st PARAM only if channel
         * has multiple PaRAMs, i.e. it can be used to submit linked transfers.
         */
        int handleSize = sizeof(ECPY_Obj) + 
                ( ((handle->assignedNumPaRams > 1) ? 1 : 0) 
                    * sizeof(IRES_EDMA3CHAN_PaRamStruct) );

        /* Allocate memory for the ECPY handle object */
        if ((ecpyHandle = myAlloc(handleSize, 0)) == NULL) {
        
            Log_print1(Diags_USER7, "[+7] ECPY_createHandle> "
                    "Could not allocate memory size 0x%x in space "
                    "IALG_EXTERNAL of type IALG_PERSIST", (IArg)handleSize);

            Log_print0(Diags_EXIT, "[+X] ECPY_createHandle> "
                "Exit (status=NULL)");

            return  (ECPY_Handle) NULL;
        }
    }

    if (handle->assignedNumPaRams > 1) {
        ecpyHandle->firstPARAMShadow =  (IRES_EDMA3CHAN_PaRamStruct *)
                ((unsigned int)ecpyHandle + sizeof(ECPY_Obj));
    }
    
    tcc     = handle->assignedTccIndices[0];
    channel = handle->assignedEdmaChannelIndex;

    /* 
     * For performance optimization: cache the IPR Register, channel-mask
     * and the ESR Register and channel bitmask in the handle.
     */
    if (tcc < 32) {
        ecpyHandle->iprRegisterAddr = (unsigned int)&_EDMA3_globalRegs->IPR;
        ecpyHandle->iprBitMask = (1UL << tcc);
        
    } else {
        ecpyHandle->iprRegisterAddr = (unsigned int)&_EDMA3_globalRegs->IPRH;
        ecpyHandle->iprBitMask = (1UL << (tcc - 32));
    }
     
    if (channel < 32) {
        ecpyHandle->esrRegisterAddr = (unsigned int)&_EDMA3_globalRegs->ESR;
        ecpyHandle->esrBitMask = (1UL << channel);

    } else {
        ecpyHandle->esrRegisterAddr = (unsigned int)&_EDMA3_globalRegs->ESRH;
        ecpyHandle->esrBitMask = (1UL << (channel - 32));
    } 
    ecpyHandle->iresHandle = handle;
    ecpyHandle->finalTransferNo = handle->assignedNumPaRams;

    /* Reset transferPending indicating initial state of channel. */
    ecpyHandle->transferPending = 0;

    Log_print0(Diags_EXIT, "[+E] ECPY_createHandle> Exit ");

    return ecpyHandle;
}

/*
 * ECPY_deleteHandle(ECPY_Handle handle)
 */
Void ECPY_deleteHandle(ECPY_Handle ecpyHandle)
{
    int handleSize;

    IRES_EDMA3CHAN2_Handle iresHandle = NULL;

    Log_print1(Diags_ENTRY, "[+E] ECPY_deleteHandle> Enter "
            "(handle=0x%x)", (IArg)ecpyHandle);

#ifdef _DEBUG_
    Assert_isTrue(ecpyHandle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    iresHandle = ecpyHandle->iresHandle;

    if ( (ecpyHandle != NULL) && (iresHandle->shadowPaRams == NULL)) {

        /* 
         * Free ecpyHandle memory ONLY IF the channel (iresHandle) has NOT
         * been assigned shadow PARAM memory, because if present, that memory
         * gets used by ECPY_createHandle to construct the ECPY handle and
         * nothing needs to be freed here.
         */
        handleSize = sizeof(ECPY_Obj) + 
                ( ((iresHandle->assignedNumPaRams > 1) ? 1 : 0) 
                   * sizeof(IRES_EDMA3CHAN_PaRamStruct) );

        /* Free allocates memory for the ECPY handle object */
        myFree(ecpyHandle, handleSize);
    }

    Log_print0(Diags_EXIT, "[+E] ECPY_deleteHandle> Exit ");
}

/*
 * ECPY_activate(ECPY_Handle handle)
 */
Void ECPY_activate(ECPY_Handle handle)
{
    unsigned int channel;
    IRES_EDMA3CHAN2_Handle iresHandle = NULL;

    Log_print1(Diags_ENTRY, "[+E] ECPY_activate> Enter "
            "(handle=0x%x)", (IArg)handle);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    iresHandle = (IRES_EDMA3CHAN2_Handle)(handle->iresHandle);
    channel    = iresHandle->assignedEdmaChannelIndex;
    
    /* Map EDMA3 Channel to the PARAM associated with channel    */ 
    if (channel < 64) {
        _EDMA3_globalRegs->DCHMAP[channel] = 
                iresHandle->assignedPaRamAddresses[0];
    }
    
    /* Enable EDMA3 Channel */
    if (channel < 32) {
        *(volatile Uns *)(&_EDMA3_globalRegs->EESR) = (1UL << channel); 

    } else if (channel < 64) {
        *(volatile Uns *)(&_EDMA3_globalRegs->EESRH) = (1UL << (channel-32)); 
    } 

    Log_print0(Diags_EXIT, "[+E] ECPY_activate> Exit ");
    
    return;
}

/*
 * ECPY_deactivate(ECPY_Handle handle)
 * 
 * Current implementation does not disable the EDMA3 channels associated with 
 * the handle to avoid race conditions involving shared EDMA channels.
 */
Void ECPY_deactivate(ECPY_Handle handle)
{
    Log_print1(Diags_ENTRY, "[+E] ECPY_deactivate> Enter "
            "(handle=0x%x)", (IArg)handle);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    Log_print0(Diags_EXIT, "[+E] ECPY_deactivate> Exit ");
}

/**
 *  @brief      Indicate the memory space that DMA transfers submitted to
 *              the channel associated with the handle will modify.
 *              This information can be used for channel to TC Queue mapping
 *              for avoiding known Silicon GEM lockup issues or for 
 *              performance improvements.
 *
 *  @param[in]  handle  ECPY channel handle.
 *  @param[in]  mode    DMA destination description of type 
 *                      IRES_EDMA3CHAN_DmaDestType.
 */
Void ECPY_setDestinationMemoryMode(ECPY_Handle handle, short mode)
{
    volatile unsigned int *dmaqnumRegister; 
    unsigned int *dmaqnum0RegisterAddr = 
           (unsigned int *)&_EDMA3_globalRegs->DMAQNUM[0];

    unsigned int edmaChannelNo = handle->iresHandle->assignedEdmaChannelIndex;
    unsigned int queueNum = 0;
   
    Log_print2(Diags_ENTRY, "[+E] ECPY_setDestinationMemoryMode> Enter "
            "(handle=0x%x) (mode=%d)", (IArg)handle, (IArg)mode);

#ifdef _DEBUG_
    Assert_isTrue(handle != NULL, (Assert_Id)NULL);
#endif /* _DEBUG_ defined only for debug profile */

    /* map the channel to EDMA Queue */

    /*  select the DMAQNUM[i] register, each register mapping 8 channels */
    dmaqnumRegister = dmaqnum0RegisterAddr + (edmaChannelNo >> 3);

    /* get configuration supplied Queue Number using the IRES_EDMA3CHAN2 */
    queueNum = handle->iresHandle->queueMap[mode];

    /* update the 3 bits corr to the Channel with the Queue number */
    (*dmaqnumRegister) |= (queueNum << (4 * (edmaChannelNo & 0x7)));

    Log_print2(Diags_USER4, "[+4] ECPY_setDestinationMemoryMode> Map "
            " EDMA Channel: %d to TC Queue: 0x%x",
            (IArg)(edmaChannelNo), (IArg)(queueNum));

    Log_print0(Diags_EXIT, "[+E] ECPY_setDestinationMemoryMode> Exit ");
}


/**
 *  @brief              Initialize the ECPY module.  Called by framework.
 */
Void ECPY_init(Void);

/**
 *  @brief              Finalization of the ECPY module.
 */
Void ECPY_exit(Void);


/*@}*/

#ifdef __cplusplus
}
#endif /* extern "C" */

/*
 *  @(#) ti.sdo.fc.ecpy; 1, 0, 0,54; 11-2-2010 18:25:30; /db/atree/library/trees/fc/fc-n22x/src/ xlibrary

 */

