/* 
 * Copyright (c) 2010, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 */
/*
 *  ======== iresman_vicp2.c ========
 */

/* This difine must precede inclusion of any xdc header files */
#define Registry_CURDESC ti_sdo_fc_ires_vicp2_desc

#include <xdc/std.h>
#include <xdc/runtime/Assert.h>
#include <xdc/runtime/Diags.h>
#include <xdc/runtime/Log.h>
#include <xdc/runtime/Registry.h>
#include <xdc/runtime/IGateProvider.h>
#include <xdc/runtime/knl/GateProcess.h>
#include <ti/sdo/fc/utils/dbc.h>

#ifdef xdc_target__os_Linux
#include <sys/mman.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <ti/sdo/fc/utils/shm.h>
#endif

#include <ti/sdo/fc/utils/osalsupport/OsalSupport.h>

#include <ti/sdo/fc/global/FCSettings.h>

#define ti_sdo_fc_ires_NOPROTOCOLREV
#include "ires_vicp2.h"
#include "iresman_vicp2.h"
#include "_vicp2.h"
#include "iresman_protocol_vicp2.h"
#include <ti/sdo/fc/utils/smgrmp.h>


#define IRESMAN_VICP2_RESOURCEFREE 0xFFFF

#define VICP2_NUM_GROUPS 20
#define VICP2_RESSIZE 0x200
#define VICP2_NUMRES ((VICP2_BUFSIZE)/(VICP2_RESSIZE))

typedef enum IRES_VICP2_RESOURCETYPE {
    IMX0,
    IMX1,
    MJCP,
    NSF
} IRES_VICP2_RESOURCETYPE;

/*
 * Holds the internal state of an initialized object
 */
typedef struct IRESMAN_Vicp2_InternalState {

    Int IMX0busy;
    Int IMX1busy;
    Int MJCPbusy;
    Int NSFbusy;
                                  /* Keeps track of busy status of resource */
    short refCount[4];
                                  /* Reference count of number of algs using
                                     the same resource */
    Int numOpens;
    Int sharedMemId;
    Void * base;
} IRESMAN_Vicp2_InternalState;

unsigned int _VICP2_buffOffsets[IRES_VICP2_NUMBUFFERS];
unsigned int _VICP2_buffLens[IRES_VICP2_NUMBUFFERS];

unsigned int _VICP2_regSize;
unsigned int _VICP2_bufSize;

/*
 * Key to obtain shared memory for this module
 */
extern Int32 _VICP2_ipcKeyBase;
/*
 * Linuxutils module for VICP2 */
extern int VICP_init();
extern int VICP_exit();


/*
 * Represents the IRES EDMA3 Resource manager
 */
static IRESMAN_Vicp2_InternalState * _resmanInternalState = NULL;

/*
 * Memory alloc Function
 */
static IRESMAN_PersistentAllocFxn * _allocFxn = NULL;

/*
 * Memory free Function
 */
static IRESMAN_PersistentFreeFxn * _freeFxn = NULL;

#ifndef xdc_target__os_Linux

/*
 * For non-Linux, non-MP case, use statically allocated memory
 */
static IRESMAN_Vicp2_InternalState _VICP2_internalState;
#endif

static IGateProvider_Handle VICP2_gate = NULL;

static SMGRMP_Handle smgr = NULL;

static void getInternalState();
static void freeInternalState();
static void initializeBufferOffsets();

Registry_Desc ti_sdo_fc_ires_vicp2_desc;

static Int regInit = 0;  /* Registry_addModule() called */
/**
 * Function to return the name of the protocol
 */
String IRESMAN_VICP2_getProtocolName()
{
    if (regInit) { 
    Log_print0(Diags_ENTRY, "[+E] IRESMAN_VICP2_getProtocolName> Enter");

    Log_print1(Diags_EXIT, "[+X] IRESMAN_VICP2_getProtocolName> Exit (name=%s)",
            (IArg)(IRESMAN_VICP2_CONSTRUCTFXNS.getName()));
    }

    return (IRESMAN_VICP2_CONSTRUCTFXNS.getName());
}

/**
 * Function to return the revision of the protocol
 */
IRES_ProtocolRevision * IRESMAN_VICP2_getProtocolRevision()
{
    IRES_ProtocolRevision * version;

    if (regInit)
    Log_print0(Diags_ENTRY, "[+E] IRESMAN_VICP2_getProtocolRevision> Enter");

    version = IRESMAN_VICP2_CONSTRUCTFXNS.getRevision();

    if (regInit)
    Log_print3(Diags_EXIT,
            "[+X] IRESMAN_VICP2_getProtocolRevision> Exit (version=(%d.%d.%d))",
            (IArg)(version->Major), (IArg)(version->Source),
            (IArg)(version->Radius));

    return (version);
}

/**
 * Function to initialize the device specific resource manager implementation
 */
IRES_Status IRESMAN_VICP2_init(IRESMAN_Params * initArgs)
/* ARGSUSED - this line tells the compiler not to warn about unused args. */
{
    SMGRMP_Attrs attrs;
    Int numResources, resourceId = -1;
    Bool smg_status = FALSE;
    Registry_Result result;
    IArg key;

    IRESMAN_VicpParams * resmanArgs = (IRESMAN_VicpParams *)initArgs;

    if (!regInit) {
        /*
         *  Enable logging. We only want to do this once, since there is
         *  no Registry_removeModule().
         */
        result = Registry_addModule(&ti_sdo_fc_ires_vicp2_desc,
                VICP2_MODNAME);

        Assert_isTrue(result == Registry_SUCCESS, (Assert_Id)NULL);

        if (result == Registry_SUCCESS) {
            /* Set the diags mask to the FC default */
            FCSettings_init();
            FCSettings_setDiags(VICP2_MODNAME);
        }
        regInit = 1;
    }

    Log_print1(Diags_ENTRY, "[+E] IRESMAN_VICP2_init> Enter (initArgs=0x%x)",
            (IArg)initArgs);

    Assert_isTrue(initArgs != NULL, (Assert_Id)NULL);

    if (VICP2_gate == NULL) {
        /* Create a gate */
        VICP2_gate = OsalSupport_createGate(_VICP2_LOCKID);
        if (VICP2_gate == NULL) {
            Log_print0(Diags_USER7, "[+7] IRESMAN_VICP2_init> Failed to "
                    "create gate!");
            return (IRES_ENOMEM);
        }
    }

    getInternalState();

    if (NULL == _resmanInternalState) {

        Log_print0(Diags_USER7,
                "[+7] IRESMAN_VICP2_init> Failed to obtain Internal state Object");

        Log_print0(Diags_EXIT, "[+X] IRESMAN_VICP2_init> Exit (status="
                "IRES_EFAIL)");

        return (IRES_EFAIL);
    }

    /*
     * Information regarding the memory allocation/free functions
     */
    _allocFxn = resmanArgs->baseConfig.allocFxn;

    _freeFxn = resmanArgs->baseConfig.freeFxn;

    key = IGateProvider_enter(VICP2_gate);
    /*
     *  Check if already initialized
     */
    VICP_init();


    initializeBufferOffsets();

    _resmanInternalState->numOpens++;

    if (NULL == smgr) {

        attrs.numScratchGroups = VICP2_NUM_GROUPS;
        attrs.numResources = (VICP2_NUMRES);
                                     /* One resource corresponds to a 1/2 K
                                        chunk of memory (0x200), Manage memory
                                        in chunks of 0x200 */
        attrs.gate = (GateProcess_Handle)VICP2_gate;

        attrs.key = (void *)_VICP2_MEMID;    /* A key specific to the resource
                                                being managed */
        smgr = SMGRMP_create(&attrs);

        if (NULL == smgr) {

            IGateProvider_leave(VICP2_gate,key);
            Log_print0(Diags_USER7, "[+7] IRESMAN_VICP2_init> Error creating"
                    " scratch resource manager.");

            Log_print0(Diags_EXIT, "[+X] IRESMAN_VICP2_init> Exit (status="
                    "IRES_EFAIL)");

            return (IRES_EFAIL);
        }

    }

    IGateProvider_leave(VICP2_gate, key);

    if ((_resmanInternalState->numOpens == 1) && (_VICP2_dm365)) {

/* TODO: */
        /* Managing VICP2_BUFSIZE of buffer space, need to pre-allocate
           the RSVD blocks */
        Log_print0(Diags_USER4, "[+4] IRESMAN_VICP2_init> Pre-allocating "
                   "the RSVD blocks");

        resourceId = (RSVD_BIMXREF_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_BIMXREF_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_PREDMEM_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_PREDMEM_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_MVMEM_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_MVMEM_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_SEQBUF1_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_SEQBUF1_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_SEQBUF2_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_SEQBUF2_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_SEQBUF3_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_SEQBUF3_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }

        resourceId = (RSVD_SEQDMEM_BASE)/(VICP2_RESSIZE);
        Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
        numResources = (RSVD_SEQDMEM_LEN/VICP2_RESSIZE);

        smg_status = SMGRMP_alloc(smgr, NULL, &resourceId, numResources, -1);

        if (FALSE == smg_status) {
            goto allocFail;
        }
    }

    Log_print0(Diags_EXIT, "[+X] IRESMAN_VICP2_init> Exit (status=IRES_OK)");

    return (IRES_OK);

allocFail:

    Log_print0(Diags_USER7,
            "[+7] IRESMAN_VICP2_init>Failed to lock the RSVD buffers");

    Log_print0(Diags_EXIT, "[+X] IRESMAN_VICP2_init> Exit (status=IRES_EFAIL)");

    return (IRES_EFAIL);
}

/**
 * Function called when RESMAN is no longer needed. Use it to free memory
 * allocated etc
 */
IRES_Status IRESMAN_VICP2_exit()
{
    IArg key;
    Log_print0(Diags_ENTRY, "[+E] IRESMAN_VICP2_exit> Enter");

    Assert_isTrue(_resmanInternalState != NULL , (Assert_Id)NULL);

    key = IGateProvider_enter(VICP2_gate);
    _resmanInternalState->numOpens--;

    freeInternalState();

    _resmanInternalState = NULL;
    IGateProvider_leave(VICP2_gate,key);

    SMGRMP_delete(smgr);

    if (VICP2_gate != NULL) {
        /* Free the gate */
        OsalSupport_deleteGate(VICP2_gate);
        VICP2_gate = NULL;
    }

    Log_print0(Diags_EXIT, "[+X] IRESMAN_VICP2_exit> Exit (status=IRES_OK)");

    return (IRES_OK);
}

/*
 *  Function that returns the IRES_Handle to the resource requested using the
 *  IRES_ResourceProtocolArgs
 */
/* ARGSUSED - this line tells the compiler not to warn about unused args. */
IRES_Handle IRESMAN_VICP2_getHandles(IALG_Handle algHandle,
        IRES_ResourceDescriptor * resDesc, Int scratchGroupId,
        IRES_Status * status)
{
    IALG_MemRec protocolMemRecs[IRES_VICP2_MAXRESOURCES];
    IALG_MemRec constructHandleMemRec;
    IRESMAN_VICP2_ConstructHandleArgs * constructArgs = NULL;
    IRES_ProtocolArgs * protocolArgs = resDesc->protocolArgs;
    IRES_Handle resourceHandle = NULL;
    Int numMemRecs = 0;
    Int i, numResources, resourceId = -1;
    Int toFreeBufIndex = -1;
    Int toFreeMemIndex = -1;
    unsigned int mem_beg = 0x0;
    unsigned int mem_end = 0x0;
    unsigned int buf_beg = 0x0;
    unsigned int buf_end = 0x0;
    unsigned int buf_num = 0x0;
    Int numBuffSwitches = 0;
    Bool smg_status = FALSE;

    /*
     * Cast the protocolArgs
     */
    IRES_VICP2_ProtocolArgs * configArgs =
            (IRES_VICP2_ProtocolArgs *) protocolArgs;

    Assert_isTrue(configArgs, (Assert_Id)NULL);

    if (_resmanInternalState == NULL) {

        Log_print0(Diags_USER7,
                "[+7] _IRESMAN_VICP2_getHandles> RMAN_register call has not "
                "happened successfully\n. Please register VICP2 resource "
                "with RMAN before requesting or releasing resources.");

        *status =  IRES_ENORESOURCE;

        return (NULL);
    }

    /*
     * Get IRES_VICP2 handle's memory requirements
     */
    numMemRecs = IRESMAN_VICP2_CONSTRUCTFXNS.getNumMemRecs
         ( (IRES_ProtocolArgs *)protocolArgs);

    Assert_isTrue(numMemRecs <= IRES_VICP2_MAXRESOURCES, (Assert_Id)NULL);

    /*
     * Allocate memory for the constructHandle Arguments
     */
    constructHandleMemRec.size = sizeof(IRESMAN_VICP2_ConstructHandleArgs);
    constructHandleMemRec.alignment = 4;
    constructHandleMemRec.space = IALG_EXTERNAL;
    constructHandleMemRec.attrs = IALG_PERSIST;

    if (FALSE == _allocFxn(&constructHandleMemRec, 1)) {
        *status =  IRES_ENOMEM;
        return (NULL);
    }

    Assert_isTrue(IALG_EXTERNAL == constructHandleMemRec.space,
            (Assert_Id)NULL);
    constructArgs = constructHandleMemRec.base;
    (constructArgs->args).size = sizeof(IRESMAN_VICP2_ConstructHandleArgs);

    if (configArgs->mode == IRES_PERSISTENT) {
        scratchGroupId = -1;
    } 
    
    if (-1 == scratchGroupId) {
        constructArgs->persistent = TRUE;
    }
    else {
        constructArgs->persistent = FALSE;
    }

    if ( configArgs->requestIMX0 ) {
        if (_resmanInternalState->IMX0busy == IRESMAN_VICP2_RESOURCEFREE){
            _resmanInternalState->IMX0busy = scratchGroupId;
            _resmanInternalState->refCount[IMX0] = 1;
            constructArgs->assignedIMX0 = 1;
        }
        else if (( -1 != _resmanInternalState->IMX0busy) &&
                (_resmanInternalState->IMX0busy == scratchGroupId)) {
            _resmanInternalState->refCount[IMX0]++;
            constructArgs->assignedIMX0 = 1;
        }
        else {
           *status = IRES_ENORESOURCE;
            constructArgs->assignedIMX0 = 0;
            goto getHandleEnd;
        }
    }

    /* DM355 doesn't have an IMX1 */
    if (_VICP2_dm365) {

        if ( configArgs->requestIMX1 ) {
            if (_resmanInternalState->IMX1busy == IRESMAN_VICP2_RESOURCEFREE){
                _resmanInternalState->IMX1busy = scratchGroupId;
                _resmanInternalState->refCount[IMX1] = 1;
                constructArgs->assignedIMX1 = 1;
            }
            else if (( -1 != _resmanInternalState->IMX1busy) &&
                    (_resmanInternalState->IMX1busy == scratchGroupId)) {
                _resmanInternalState->refCount[IMX1]++;
                constructArgs->assignedIMX1 = 1;
            }
            else {
               *status = IRES_ENORESOURCE;
                constructArgs->assignedIMX1 = 0;
                goto getHandleEnd;
            }
        }
    }

    if (configArgs->requestMJCP) {
        if (_resmanInternalState->MJCPbusy == IRESMAN_VICP2_RESOURCEFREE) {
            _resmanInternalState->MJCPbusy = scratchGroupId;
            _resmanInternalState->refCount[MJCP] = 1;
            constructArgs->assignedMJCP = 1;
        }
        else if (( -1 != _resmanInternalState->MJCPbusy) &&
                (_resmanInternalState->MJCPbusy == scratchGroupId)) {
            _resmanInternalState->refCount[MJCP]++;
            constructArgs->assignedMJCP = 1;
        }
        else {
           *status = IRES_ENORESOURCE;
            constructArgs->assignedMJCP = 0;
            goto getHandleEnd;
        }
    }

    /* Ignore requests for NSF on the DM355 */
    if (_VICP2_dm365) {

        if (configArgs->requestNSF) {
            if (_resmanInternalState->NSFbusy == IRESMAN_VICP2_RESOURCEFREE) {
                _resmanInternalState->NSFbusy = scratchGroupId;
                _resmanInternalState->refCount[NSF] = 1;
                constructArgs->assignedNSF = 1;
            }
            else if (( -1 != _resmanInternalState->NSFbusy) &&
                    (_resmanInternalState->NSFbusy == scratchGroupId)) {
                _resmanInternalState->refCount[NSF]++;
                constructArgs->assignedNSF = 1;
            }
            else {
               *status = IRES_ENORESOURCE;
                constructArgs->assignedNSF = 0;
                goto getHandleEnd;
            }
        }
    }

    /* Initialize all buffers to  RESNA */
    for (i = 0; i < IRES_VICP2_NUMBUFFERS; i++) {
        constructArgs->assignedIMCOPBuffers[i] = RESNA;
    }


    /* Obtain buffers here */
    if (configArgs->requestIMCOPBuffers) {

        for (i = 0; i < IRES_VICP2_NUMBUFFERS; i++) {

            if (configArgs->requestIMCOPBuffers & (0x1 << i)) {

                /* Ignore invalid requests and continue */
                if (_VICP2_buffOffsets[i] == RESNA) {
                    continue;
                }

                /* Valid Request for Buffer i */
                /* Each resource corresponds to 0x200 or RESSIZE of memory */
                resourceId = (_VICP2_buffOffsets[i])/(VICP2_RESSIZE);
                Assert_isTrue((resourceId < VICP2_NUMRES), (Assert_Id)NULL);
                numResources = (_VICP2_buffLens[i]/VICP2_RESSIZE);

                if (TRUE == SMGRMP_alloc(smgr, algHandle, &resourceId,
                        numResources, scratchGroupId)) {

                    constructArgs->assignedIMCOPBuffers[i] =
                            _VICP2_buffOffsets[i];
                    constructArgs->assignedIMCOPBufferLens[i] =
                            _VICP2_buffLens[i];
                }
                else {
                    /* Free buffers allocated so far */
                    *status = IRES_ENORESOURCE;
                    toFreeBufIndex = i-1;
                    goto getHandleEnd;
                }
            }
        }
        toFreeBufIndex = i - 1;
    }

    constructArgs->numMemBufs = configArgs->numMemoryRequests;

    /* Request for memory chunks */
    for (i = 0; i < configArgs->numMemoryRequests; i++) {

        resourceId = -1;

        /* Round up request size to 512b */
        numResources = ((configArgs->memRequestSizes[i] +
                ((VICP2_RESSIZE) - 1))/(VICP2_RESSIZE));

        Assert_isTrue(numResources <VICP2_NUMRES, (Assert_Id)NULL);

        /* Request for Buffer i */
        if (TRUE == SMGRMP_alloc( smgr, algHandle, &resourceId, numResources,
                scratchGroupId)) {

            constructArgs->assignedMemAddrs[i] = resourceId * VICP2_RESSIZE;
            constructArgs->assignedMemSizes[i] = numResources * VICP2_RESSIZE;
            /* Also update constructArgs->assignedIMCOPBuffers[k], where
               k corresponds to one or more buffers that this block overlaps
               with. */
            mem_beg = constructArgs->assignedMemAddrs[i];
            mem_end = mem_beg + constructArgs->assignedMemSizes[i];

            buf_num = 0;
            buf_beg =  _VICP2_buffOffsets[buf_num];
            buf_end = buf_beg + _VICP2_buffLens[buf_num];

            //while ((buf_beg < mem_end) && (buf_num < IRES_VICP2_NUMBUFFERS)) {
            while (buf_num < IRES_VICP2_NUMBUFFERS) {

                //if ((buf_beg != RESNA) && (buf_end > mem_beg)) {
                if ((buf_beg != RESNA) && (buf_end > mem_beg) &&
                        (buf_beg < mem_end)) {

                    /* Update with the address of that buffer !! */
                    if (constructArgs->assignedIMCOPBuffers[buf_num] == RESNA) {

                        /* Even though these buffers weren't completely
                           assigned to this alg, these arrays are being updated
                           with their addresses, so that the alg can turn on
                           these buff switches */
                        constructArgs->assignedIMCOPBuffers[buf_num] =
                                _VICP2_buffOffsets[buf_num];

                        constructArgs->assignedIMCOPBufferLens[buf_num] =
                                _VICP2_buffLens[buf_num];

                        /* Need to keep track of these entries as when its time
                           to free buffers, these aren't freed, as
                           they were never explicitly granted to the alg */
                        constructArgs->buffSwitchesOnly[numBuffSwitches++] =
                                buf_num;

                        constructArgs->numBuffSwitches = numBuffSwitches;
                    }
                }

                buf_num++;
                buf_beg = _VICP2_buffOffsets[buf_num%IRES_VICP2_NUMBUFFERS];
                buf_end = buf_beg +
                        _VICP2_buffLens[buf_num%IRES_VICP2_NUMBUFFERS];
            }
        }
        else {
            *status = IRES_ENORESOURCE;
            toFreeMemIndex = i-1;
            goto getHandleEnd;
        }
    }
    toFreeMemIndex = i - 1;


    /*
     * Obtain memory resources to build the handle
     */
    IRESMAN_VICP2_CONSTRUCTFXNS.getMemRecs((IRES_Handle)NULL,
            (IRES_ProtocolArgs *)protocolArgs, protocolMemRecs);

   /*
    * Allocate that memory
    */
    if (FALSE == _allocFxn(protocolMemRecs,
            numMemRecs)) {
        *status = IRES_ENOMEM;
        resourceHandle = NULL;
    }
    else {

        /*
         * Use the CONSTRUCT interface of the protocol implementation to
         * construct the handle within the allocated resource
         */
        resourceHandle = IRESMAN_VICP2_CONSTRUCTFXNS.constructHandle(
                protocolArgs, protocolMemRecs,
                (IRESMAN_ConstructArgs *)constructArgs, status);
    }

    if (*status != IRES_OK) {
        /* Need to free Buffer resources here  */
    }

getHandleEnd:

    if (*status != IRES_OK) {

        /* First, fix assignedIMCOPBuffers array to remove BuffSwitches entries
           only. These entries were updated in assignedIMCOPBuffers, but
           these buffers were never explicitly assigned to the alg */
        for (i = 0; i < constructArgs->numBuffSwitches; i++) {
            constructArgs->assignedIMCOPBuffers[constructArgs->buffSwitchesOnly
                    [i]] = RESNA;
        }

        /* Free the rest of the allocated buffers */
        while (toFreeBufIndex >= 0) {

            if (constructArgs->assignedIMCOPBuffers[toFreeBufIndex] == RESNA) {
                toFreeBufIndex--;
                continue;
            }
            resourceId = (constructArgs->assignedIMCOPBuffers[toFreeBufIndex])/
                    (VICP2_RESSIZE);

            numResources = (constructArgs->assignedIMCOPBufferLens[
                    toFreeBufIndex])/(VICP2_RESSIZE);

            smg_status = SMGRMP_free(smgr, algHandle, resourceId, numResources,
                    scratchGroupId);

            Assert_isTrue(smg_status, (Assert_Id)NULL);

            if (smg_status == FALSE) {
                *status = IRES_EFAIL;
            }

            toFreeBufIndex--;
        }

        while (toFreeMemIndex >= 0) {

            resourceId = (constructArgs->assignedMemAddrs[toFreeMemIndex])/
                    (VICP2_RESSIZE);
            numResources = ((configArgs->memRequestSizes[toFreeMemIndex] +
                    ((VICP2_RESSIZE) - 1))/(VICP2_RESSIZE));
            smg_status = SMGRMP_free(smgr, algHandle, resourceId, numResources,
                    scratchGroupId);
            Assert_isTrue(smg_status, (Assert_Id)NULL);

            if (smg_status == FALSE) {
                *status = IRES_EFAIL;
            }

            toFreeMemIndex--;

       }

       if (constructArgs->assignedIMX0 == 1) {
           _resmanInternalState->refCount[IMX0]--;
            if (_resmanInternalState->refCount == 0 ) {
                _resmanInternalState->IMX0busy = IRESMAN_VICP2_RESOURCEFREE;
            }
        }
        else if (constructArgs->assignedIMX1 == 1) {
           _resmanInternalState->refCount[IMX1]--;
            if (_resmanInternalState->refCount == 0 ) {
                _resmanInternalState->IMX1busy = IRESMAN_VICP2_RESOURCEFREE;
            }
        }
        else if (constructArgs->assignedMJCP == 1) {
           _resmanInternalState->refCount[MJCP]--;
            if (_resmanInternalState->refCount == 0 ) {
                _resmanInternalState->MJCPbusy = IRESMAN_VICP2_RESOURCEFREE;
            }
        }
        else if (constructArgs->assignedNSF == 1) {
           _resmanInternalState->refCount[NSF]--;
            if (_resmanInternalState->refCount == 0 ) {
                _resmanInternalState->NSFbusy = IRESMAN_VICP2_RESOURCEFREE;
            }
        }
    }

    /*
     * Free constructArgs memory, that was just to pass the arguments
     */
    _freeFxn(&constructHandleMemRec, 1);

    resDesc->handle = (IRES_Obj *)resourceHandle;

    resDesc->revision->Major = 1;
    resDesc->revision->Source = 0;
    resDesc->revision->Radius = 0;

    return (resourceHandle);
}

/**
 *  Frees handles back to the resource pool resources to free
 */
/*  ARGSUSED - this line tells the compiler not to warn about unused args */
IRES_Status IRESMAN_VICP2_freeHandles(IALG_Handle algHandle,
        IRES_Handle algResourceHandle, IRES_ResourceDescriptor * resDesc,
        Int scratchGroupId)
{
    IALG_MemRec resourceMemRecs[IRES_VICP2_MAXRESOURCES];
    Int numMemRecs;
    Int i, numResources, resourceId = -1;
    Bool smg_status = FALSE;
    IRES_Status retStatus = IRES_OK;
    IRES_ProtocolArgs * protocolArgs = resDesc->protocolArgs;


    IRES_VICP2_Handle resourceHandle = (IRES_VICP2_Handle)algResourceHandle;

    /*
     * Cast the protocolArgs 
     */
    IRES_VICP2_ProtocolArgs * configArgs = 
            (IRES_VICP2_ProtocolArgs *) protocolArgs;

    Assert_isTrue(configArgs, (Assert_Id)NULL);
    Assert_isTrue(algResourceHandle, (Assert_Id)NULL);

    if (_resmanInternalState == NULL) {

        Log_print0(Diags_USER7,
                "[+7] _IRESMAN_VICP2_freeHandles> RMAN_register call has not "
                "happened successfully\n. Please register VICP2 resource "
                "with RMAN before requesting or releasing resources.");

        retStatus = IRES_ENORESOURCE;
        return (retStatus);
    }

    if (configArgs->mode == IRES_PERSISTENT) {
        scratchGroupId = -1;
    } 
 
    /*
     * Free the handles to the resource manager
     */
    if (resourceHandle->assignedIMX1 == 1) {
        if (_resmanInternalState->refCount[IMX1] == 0) {
            retStatus =  IRES_ENORESOURCE;
            return (retStatus);
        }
        _resmanInternalState->refCount[IMX1]--;
        if (_resmanInternalState->refCount == 0 ) {
            _resmanInternalState->IMX1busy = IRESMAN_VICP2_RESOURCEFREE;
        }
    }
    else if (resourceHandle->assignedIMX0 == 1) {
        if (_resmanInternalState->refCount[IMX0] == 0) {
            retStatus = IRES_ENORESOURCE;
            return (retStatus);
        }
        _resmanInternalState->refCount[IMX0]--;
        if (_resmanInternalState->refCount == 0 ) {
            _resmanInternalState->IMX0busy = IRESMAN_VICP2_RESOURCEFREE;
        }
    }
   else if (resourceHandle->assignedMJCP == 1) {
        if (_resmanInternalState->refCount[MJCP] == 0) {
            retStatus = IRES_ENORESOURCE;
            return (retStatus);
        }
        _resmanInternalState->refCount[MJCP]--;
        if (_resmanInternalState->refCount == 0 ) {
            _resmanInternalState->MJCPbusy = IRESMAN_VICP2_RESOURCEFREE;
        }
    }
    else if (resourceHandle->assignedNSF == 1) {
        if (_resmanInternalState->refCount[NSF] == 0) {
            retStatus = IRES_ENORESOURCE;
            return (retStatus);
        }
        _resmanInternalState->refCount[NSF]--;
        if (_resmanInternalState->refCount == 0 ) {
            _resmanInternalState->NSFbusy = IRESMAN_VICP2_RESOURCEFREE;
        }
    }


    /* Free buffers here */
    /* First, fix assignedIMCOPBuffers array to remove BuffSwitches entries
       only. These buffers were added to teh assignedIMCOPBuffer entries,
       but weren't really explicitly allocated to the alg, only memory
       chunks out of these buffers were allocated. */
    for (i = 0; i < ((_VICP2_IntObj *)(resourceHandle))->numBuffSwitches; i++) {
        resourceHandle->assignedIMCOPBuffers[((_VICP2_IntObj *)
                resourceHandle)->buffSwitchesOnly[i]] = RESNA;
    }

    for (i = 0; i < IRES_VICP2_NUMBUFFERS; i++) {

        if (RESNA != resourceHandle->assignedIMCOPBuffers[i]) {

            /* Valid Assigned Buffer i */
            /* Each resource corresponds to 0x200 of RESSIZE of memory */
            resourceId = ((resourceHandle->assignedIMCOPBuffers[i] -
                    ((_VICP2_IntObj *)resourceHandle)->configBase))/
                    (VICP2_RESSIZE);

            numResources = (_VICP2_buffLens[i])/(VICP2_RESSIZE);

            smg_status =  SMGRMP_free(smgr, algHandle, resourceId, numResources,
                    scratchGroupId);

            Assert_isTrue(smg_status, (Assert_Id)NULL);

            if (FALSE == smg_status) {
                retStatus = IRES_EFAIL;
            }
        }
    }

    for (i = 0; i < resourceHandle->numMemBufs;i++ ) {

        resourceId = (resourceHandle->assignedMemAddrs[i] -
                ((_VICP2_IntObj *)resourceHandle)->configBase)/VICP2_RESSIZE;
        numResources = ((resourceHandle->assignedMemSizes[i] +
                ((VICP2_RESSIZE) - 1))/(VICP2_RESSIZE));
        smg_status = SMGRMP_free(smgr, algHandle, resourceId, numResources,
                scratchGroupId);

        Assert_isTrue(smg_status, (Assert_Id)NULL);
   }

   /*
    * Free memory resources to build the handle
    */

    /*
     * Obtain memory resources to free and free them
     */
    numMemRecs = IRESMAN_VICP2_CONSTRUCTFXNS.getNumMemRecs
        ( (IRES_ProtocolArgs *)protocolArgs);

    IRESMAN_VICP2_CONSTRUCTFXNS.getMemRecs((IRES_Handle)NULL,
            (IRES_ProtocolArgs *) protocolArgs, resourceMemRecs);

    resourceMemRecs[0].base = algResourceHandle;

    /*
     * Use IRES_VICP2_RESOURCEPROTOCOL to de-init the resource protocol
     * if required
     */
    IRESMAN_VICP2_CONSTRUCTFXNS.destructHandle(algResourceHandle);

    /*
     * Free the memory for the handles
     */
    _freeFxn(resourceMemRecs, numMemRecs);

    return (IRES_OK);
}


/*
 *  ======= setInternalState =======
 */
static Void setInternalState(Void * addr)
{
    IRESMAN_Vicp2_InternalState * intState =
            (IRESMAN_Vicp2_InternalState *)addr;

    if (NULL != intState) {
        intState->numOpens = 0;

        intState->IMX0busy = IRESMAN_VICP2_RESOURCEFREE;
        intState->IMX1busy = IRESMAN_VICP2_RESOURCEFREE;
        intState->MJCPbusy = IRESMAN_VICP2_RESOURCEFREE;
        intState->NSFbusy = IRESMAN_VICP2_RESOURCEFREE;

        intState->refCount[0] = intState->refCount[1] =
                intState->refCount[2] = intState->refCount[3] = 0;
        intState->base = NULL;
    }
}


/*
 *  ======== getInternalState ========
 */
static void getInternalState()
{
    Int sharedMemId = -1;
#ifdef xdc_target__os_Linux

    _resmanInternalState = (IRESMAN_Vicp2_InternalState *)SHM_getObj(
            VICP2_gate, sizeof(IRESMAN_Vicp2_InternalState),
            (key_t)_VICP2_SHMID, &setInternalState, &sharedMemId);
#else
    _resmanInternalState = (IRESMAN_Vicp2_InternalState *)
            &_VICP2_internalState;
    setInternalState(_resmanInternalState);
    sharedMemId = 0;
#endif
    _resmanInternalState->sharedMemId = sharedMemId;
}

/*
 *  ======== freeInternalState ========
 */
static void freeInternalState()
{
    VICP_exit();
#ifdef xdc_target__os_Linux
   SHM_freeObj(_resmanInternalState, _resmanInternalState->sharedMemId);
#else
    _resmanInternalState = NULL;
#endif
}

static void initializeBufferOffsets() {

    _VICP2_regSize = VICP2_REGSIZE;
    _VICP2_bufSize = VICP2_BUFSIZE;

    _VICP2_buffOffsets[IMCOP_BUFFER_IMGBUF_A]           = IMGBUF_A_FULL_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMGBUF_B]           = IMGBUF_B_FULL_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMGBUF_C]           = IMGBUF_C_FULL_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMGBUF_D]           = IMGBUF_D_FULL_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMGBUF_E]           = IMGBUF_E_FULL_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMX0COEFFBUF]       = IMX0COEFFBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMX0CMDBUF]         = IMX0CMDBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMX1COEFFBUF]       = IMX1COEFFBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_IMX1CMDBUF]         = IMX1CMDBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_BABYIMXCMDBUFFER]   = BABYIMXCMDBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_BABYIMXORIGBUFFER]  = BABYIMXORIGBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_BABYIMXREFBUFFER]   = BABYIMXREFBUF_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_SEQUENCERPMEM]      = SEQ_PMEM_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_QIQMEM]             = QIQMEM_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_HUFFMEM]            = HUFFMEM_BASE;
    _VICP2_buffOffsets[IMCOP_BUFFER_SEQBUF4]            = SEQBUF4_BASE;

    _VICP2_buffLens[IMCOP_BUFFER_IMGBUF_A]              = IMGBUF_A_FULL_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMGBUF_B]              = IMGBUF_B_FULL_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMGBUF_C]              = IMGBUF_C_FULL_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMGBUF_D]              = IMGBUF_D_FULL_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMGBUF_E]              = IMGBUF_E_FULL_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMX0COEFFBUF]          = IMX0COEFFBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMX0CMDBUF]            = IMX0CMDBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMX1COEFFBUF]          = IMX1COEFFBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_IMX1CMDBUF]            = IMX1CMDBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_BABYIMXCMDBUFFER]      = BABYIMXCMDBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_BABYIMXORIGBUFFER]     = BABYIMXORIGBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_BABYIMXREFBUFFER]      = BABYIMXREFBUF_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_SEQUENCERPMEM]         = SEQ_PMEM_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_QIQMEM]                = QIQMEM_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_HUFFMEM]               = HUFFMEM_LEN;
    _VICP2_buffLens[IMCOP_BUFFER_SEQBUF4]               = SEQBUF4_LEN;
}

IRESMAN_Fxns IRESMAN_VICP2 =  {
    IRESMAN_VICP2_getProtocolName,
    IRESMAN_VICP2_getProtocolRevision,
    IRESMAN_VICP2_init,
    IRESMAN_VICP2_exit,
    IRESMAN_VICP2_getHandles,
    IRESMAN_VICP2_freeHandles
};

IRESMAN_VicpParams IRESMAN_VICP2_DEFAULTPARAM;

IRESMAN_Fxns * VICP2_getIresmanFxns()
{
    return ((Void *)(&IRESMAN_VICP2));
}


IRESMAN_Params * VICP2_getDefaultParams()
{
    IRESMAN_VICP2_DEFAULTPARAM.baseConfig.size = 
            sizeof(IRESMAN_Params);

    return ((IRESMAN_Params *)&IRESMAN_VICP2_DEFAULTPARAM);
}
/*
 *  @(#) ti.sdo.fc.ires.vicp; 1, 0, 1,255; 11-2-2010 18:26:38; /db/atree/library/trees/fc/fc-n22x/src/ xlibrary

 */

